/*****************************************************************************
 * NCSA Xmx X/Motif Extension Library.                                       *
 *                                                                           *
 * Designed and Implemented by Marc Andreessen (marca@ncsa.uiuc.edu).        *
 *                                                                           *
 * Software Development Group                                                *
 * National Center for Supercomputing Applications                           *
 * University of Illinois at Urbana-Champaign.                               *
 *                                                                           *
 * This is BETA release software.  As such it may contain software bugs and  *
 * exhibit inconsistencies.                                                  *
 *                                                                           *
 * Copyright (c) 1992 The Board of Trustees of the University of Illinois.   *
 *                                                                           *
 * Permission to use, copy, and modify this software and its                 *
 * documentation for educational, research, and non-profit purposes is       *
 * hereby granted, provided that the above copyright notice, the original    *
 * authors names, and this permission notice appear in all such copies.      *
 * Any distribution of this software requires the explicit and written       *
 * authorization of the authors.                                             *
 *                                                                           *
 * The University of Illinois makes no representations about the             *
 * suitability of this software for any purpose.  It is provided "as is"     *
 * without warranty of any kind.                                             *
 *****************************************************************************/

/* $Header: /usr3/people/gbourhis/pv3/new/RCS/Xmx2.c,v 1.1 92/09/18 10:55:26 marca Exp $ */

#ifdef RCSLOG
$Log:	Xmx2.c,v $
 * Revision 1.1  92/09/18  10:55:26  marca
 * Initial revision
 * 
#endif

#include "XmxP.h"

/* ------------------------------------------------------------------------ */
/* --------------------------- PRIVATE ROUTINES --------------------------- */
/* ------------------------------------------------------------------------ */

/* ----------------------- _XmxMenuAddEntryToRecord ----------------------- */

/* Create a new MenuEntry and add it to the head of a MenuRecord list. */
private void 
_XmxMenuAddEntryToRecord (XmxMenuRecord *rec, Widget w, int token)
{
  XmxMenuEntry *_ent;

  /* Create new menu entry. */
  _ent = (XmxMenuEntry *)malloc (sizeof (XmxMenuEntry));
  _ent->w = w;
  _ent->token = token;

  /* Add rest of list to tail of this entry. */
  _ent->next = rec->first_entry;

  /* Make this entry head of list. */
  rec->first_entry = _ent;

  return;
}

/* ---------------------- _XmxMenuGetEntryFromRecord ---------------------- */

private XmxMenuEntry *
_XmxMenuGetEntryFromRecord (XmxMenuRecord *rec, int token)
{
  /* Given token, fetch the corresponding entry. */
  XmxMenuEntry *_ent;
  int _done;

  /* Search the linked list. */
  _ent = rec->first_entry;
  _done = 0;
  while (_ent != NULL && !_done)
    {
      if (_ent->token == token)
        _done = 1;
      else
        _ent = _ent->next;
    }

  /* Punish the application for asking for a nonexistent entry. */
  assert (_done);
  return _ent;
}

/* ------------------------- _XmxMenuCreateRecord ------------------------- */

/* Create a new MenuRecord and clear out its list. */
private XmxMenuRecord *
_XmxMenuCreateRecord (Widget base)
{
  XmxMenuRecord *_rec;

  /* Create the new XmxMenuRecord. */
  _rec = (XmxMenuRecord *)malloc (sizeof (XmxMenuRecord));
  _rec->base = base;
  _rec->first_entry = NULL;

  return _rec;
}

/* ------------------------------------------------------------------------ */
/* --------------------------- PUBLIC ROUTINES ---------------------------- */
/* ------------------------------------------------------------------------ */

/* --------------------------- XmxRSetSensitive --------------------------- */

/* args NOT used on Widget */
public void
XmxRSetSensitive (XmxMenuRecord *rec, int token, int state)
{
  XmxMenuEntry *_entry;

  assert (state == XmxSensitive || state == XmxUnsensitive);
  _entry = _XmxMenuGetEntryFromRecord (rec, XmxExtractToken (token));
  /* XtSetSensitive propagates down Widget hierarchy. */
  XtSetSensitive (_entry->w, (state == XmxSensitive) ? True : False);
  
  return;
}

/* -------------------------- XmxRSetToggleState -------------------------- */

/* args not used */
public void
XmxRSetToggleState (XmxMenuRecord *rec, int token, int state)
{
  XmxMenuEntry *_entry;

  assert (state == XmxSet || state == XmxUnset);
  _entry = _XmxMenuGetEntryFromRecord (rec, XmxExtractToken (token));
  XmToggleButtonGadgetSetState 
    (_entry->w, (state == XmxSet) ? True : False, False);

  return;
}

/* ------------------------- XmxRUnsetAllToggles -------------------------- */

/* args not used */
public void
XmxRUnsetAllToggles (XmxMenuRecord *rec)
{
  XmxMenuEntry *_ent;

  for (_ent = rec->first_entry; _ent != NULL; _ent = _ent->next)
    XmToggleButtonGadgetSetState
      (_ent->w, False, False);

  return;
}

/* ----------------------- XmxRSetOptionMenuHistory ----------------------- */

/* args used on Widget */
public void
XmxRSetOptionMenuHistory (XmxMenuRecord *rec, int token)
{
  XmxMenuEntry *_entry;

  _entry = _XmxMenuGetEntryFromRecord (rec, XmxExtractToken (token));
  XmxSetArg (XmNmenuHistory, _entry->w);
  XtSetValues (rec->base, Xmx_wargs, Xmx_n);

  Xmx_n = 0;
  return;
}

/* ---------------------------- XmxRSetValues ----------------------------- */

/* args used on Widget */
public void
XmxRSetValues (XmxMenuRecord *rec, int token)
{
  XmxMenuEntry *_entry;

  /* Strip out uniqid, if present. */
  _entry = _XmxMenuGetEntryFromRecord (rec, XmxExtractToken (token));
  XtSetValues (_entry->w, Xmx_wargs, Xmx_n);

  Xmx_n = 0;
  return;
}

/* ---------------------------- XmxRGetWidget ----------------------------- */

/* args irrelevant */
public Widget
XmxRGetWidget (XmxMenuRecord *rec, int token)
{
  XmxMenuEntry *_entry;

  /* Strip out uniqid, if present. */
  _entry = _XmxMenuGetEntryFromRecord (rec, XmxExtractToken (token));

  return _entry->w;
}

/* -------------------------- XmxRMakeOptionMenu -------------------------- */

/* args apply to pulldown menu */
public XmxMenuRecord *
XmxRMakeOptionMenu (Widget parent, String name, XtCallbackProc cb,
                    XmxOptionMenuStruct *opts)
{
  XmxMenuRecord *_rec;
  Widget _pulldown, _button, _menuhist = (Widget)NULL;
  int _i;

  /* Create a pulldown menupane to attach to the option menu;
     preloaded wargs affect this. */
  _pulldown = XmCreatePulldownMenu (parent, "pulldownmenu", Xmx_wargs, Xmx_n);

  /* menuHistory will not be applied to _pulldown, so we'll modify
     _rec directly after creating the option menu. */
  _rec = _XmxMenuCreateRecord (_pulldown);

  /* Create pushbutton gadgets as childen of the pulldown menu. */
  _i = 0;
  while (opts[_i].namestr)
    {
      Xmx_n = 0;
      XmxSetArg (XmNlabelString,
                 XmStringCreateLtoR (opts[_i].namestr,
                                     XmSTRING_DEFAULT_CHARSET));
      _button = XmCreatePushButtonGadget (_pulldown, "pushbutton",
                                          Xmx_wargs, Xmx_n);
      XtManageChild (_button);
      XtAddCallback (_button, XmNactivateCallback, cb,
                     _XmxMakeClientData (opts[_i].data));
      if (opts[_i].set_state == XmxSet)
        _menuhist = _button;

      _XmxMenuAddEntryToRecord (_rec, _button, opts[_i].data);

      _i++;
    }

  /* Create the option menu itself; tie in the pulldown menu. */
  Xmx_n = 0;
  XmxSetArg (XmNsubMenuId, _pulldown);
  if (_menuhist != (Widget)NULL)
    XmxSetArg (XmNmenuHistory, _menuhist);
  Xmx_w = XmCreateOptionMenu (parent, "optionmenu", Xmx_wargs, Xmx_n);
  XtManageChild (Xmx_w);

  XmxSetArg (XmNalignment, XmALIGNMENT_BEGINNING);
  XmxSetValues (XmOptionButtonGadget (Xmx_w));

  if (name)
    {
      XmxSetArg (XmNlabelString,
                 XmStringCreateLtoR (name, XmSTRING_DEFAULT_CHARSET));
      XmxSetValues (XmOptionLabelGadget (Xmx_w));
    }

  /* Explicitly set base Widget of record. */
  _rec->base = Xmx_w;

  Xmx_n = 0;
  return _rec;
}

/* -------------------------- XmxRMakeToggleMenu -------------------------- */

/* args apply to radiobox or optionbox */
public XmxMenuRecord *
XmxRMakeToggleMenu (Widget parent, int behavior, XtCallbackProc cb,
                    XmxToggleMenuStruct *opts)
{
  XmxMenuRecord *_rec;
  Widget _box;
  int _i;

  assert (behavior == XmxOneOfMany || behavior == XmxNOfMany);
  switch (behavior)
    {
    case XmxOneOfMany:
      _box = XmxMakeRadioBox (parent);
      break;
    case XmxNOfMany:
      _box = XmxMakeOptionBox (parent);
      break;
    }

  _rec = _XmxMenuCreateRecord (_box);

  _i = 0;
  while (opts[_i].namestr)
    {
      XmxMakeToggleButton (_box, opts[_i].namestr, cb, opts[_i].data);
      XmxSetToggleButton (Xmx_w, opts[_i].set_state);
      _XmxMenuAddEntryToRecord (_rec, Xmx_w, opts[_i].data);

      _i++;
    }

  Xmx_w = _box;

  Xmx_n = 0;
  return _rec;
}

/* -------------------------- _XmxRCreateMenubar -------------------------- */

/* Possible deficiency: will not be able to grey out a submenu
   (cascade button). */
private void 
_XmxRCreateMenubar (Widget menu, XmxMenubarStruct *menulist, 
                    XmxMenuRecord *rec)
{
  int _i;
  Widget *_buttons;
  int _separators = 0, _nitems;

  _nitems = 0;
  while (menulist[_nitems].namestr)
    _nitems++;

  _buttons = (Widget *)XtMalloc (_nitems * sizeof (Widget));

  for (_i = 0; _i < _nitems; _i++)
    {
      /* Name of "----" means make a separator. */
      if (strcmp(menulist[_i].namestr, "----") == 0)
        {
          XtCreateManagedWidget ("separator", xmSeparatorGadgetClass,
                                 menu, NULL, 0);
          _separators++;
        }
      /* A function means it's an ordinary entry with callback. */
      else if (menulist[_i].func)
        {
          Xmx_n = 0;
          if (menulist[_i].mnemonic)
            XmxSetArg (XmNmnemonic, menulist[_i].mnemonic);
          if (menulist[_i].namestr[0] == '#' ||
              menulist[_i].namestr[0] == '<') /* option/toggle button */
            {
              /* A toggle button is diamond-shaped. */
              if (menulist[_i].namestr[0] == '<')
                XmxSetArg (XmNindicatorType, XmONE_OF_MANY);

	      /* Make sure the button shows up even when toggled off. */
              if (menulist[_i].namestr[0] == '#')
                XmxSetArg (XmNvisibleWhenOff, True);

              /* Ignore first character of label. */
              XmxSetArg (XmNlabelString, XmxMakeXmstrFromString
                         (&(menulist[_i].namestr[1])));
              _buttons[_i - _separators] = XtCreateManagedWidget
                ("togglebutton", xmToggleButtonGadgetClass,
                 menu, Xmx_wargs, Xmx_n);
              XtAddCallback
                (_buttons[_i - _separators], XmNvalueChangedCallback,
                 menulist[_i].func, _XmxMakeClientData (menulist[_i].data));

              /* Add thie button to the menu record. */
              _XmxMenuAddEntryToRecord 
                (rec, _buttons[_i - _separators], menulist[_i].data);
            }
          else /* regular button */
            {
              XmxSetArg (XmNlabelString, XmStringCreateLtoR
                         (menulist[_i].namestr, XmSTRING_DEFAULT_CHARSET));
              _buttons[_i - _separators] = XtCreateManagedWidget
                ("pushbutton", xmPushButtonGadgetClass,
                 menu, Xmx_wargs, Xmx_n);
              XtAddCallback 
                (_buttons[_i - _separators], XmNactivateCallback,
                 menulist[_i].func, _XmxMakeClientData (menulist[_i].data));

              /* Add thie button to the menu record. */
              _XmxMenuAddEntryToRecord 
                (rec, _buttons[_i - _separators], menulist[_i].data);
            }
        }

      /* No function and no submenu entry means it's just a label. */
      else if (menulist[_i].sub_menu == (XmxMenubarStruct *)NULL)
        {
          Xmx_n = 0;
          XmxSetArg (XmNlabelString, XmStringCreateLtoR
                     (menulist[_i].namestr, XmSTRING_DEFAULT_CHARSET));
          _buttons[_i - _separators] = XtCreateManagedWidget
            ("label", xmLabelGadgetClass, menu, Xmx_wargs, Xmx_n);
        }

      /* If all else fails, it's a submenu. */
      else
        {
          Widget _sub_menu;
          _sub_menu = XmCreatePulldownMenu (menu, "pulldownmenu", NULL, 0);

          Xmx_n = 0;
          XmxSetArg (XmNsubMenuId, _sub_menu);
          if (menulist[_i].mnemonic)
            XmxSetArg (XmNmnemonic, menulist[_i].mnemonic);
          XmxSetArg (XmNlabelString, XmStringCreateLtoR
                     (menulist[_i].namestr, XmSTRING_DEFAULT_CHARSET));
          _buttons[_i - _separators] = XtCreateWidget
            ("cascadebutton", xmCascadeButtonGadgetClass,
             menu, Xmx_wargs, Xmx_n);

          /* If name is "Help", put on far right. */
          if (strcmp (menulist[_i].namestr, "Help") == 0)
            {
              Xmx_n = 0;
              XmxSetArg (XmNmenuHelpWidget, _buttons[_i - _separators]);
              XtSetValues (menu, Xmx_wargs, Xmx_n);
            }
          /* Recursively create new submenu. */
          _XmxRCreateMenubar (_sub_menu, menulist[_i].sub_menu, rec);
        }
    }

  XtManageChildren (_buttons, _nitems - _separators);

  return;
}

/* --------------------------- XmxRMakeMenubar ---------------------------- */

/* args apply to menubar */
public XmxMenuRecord *
XmxRMakeMenubar (Widget parent, XmxMenubarStruct *mainmenu)
{
  Widget _menubar;
  XmxMenuRecord *_rec;

  /* Preset resources applied to main menubar only. */
  _menubar = XmCreateMenuBar (parent, "menubar", Xmx_wargs, Xmx_n);
  XtManageChild (_menubar);

  /* Create the new XmxMenuRecord. */
  _rec = _XmxMenuCreateRecord (_menubar);

  Xmx_n = 0;
  _XmxRCreateMenubar (_menubar, mainmenu, _rec);

  Xmx_n = 0;
  Xmx_w = _menubar;
  return _rec;
}
