/*****************************************************************************
 * NCSA Polyview 3.0                                                         *
 *                                                                           *
 * Version 3 changes and additions by Marc Andreessen.                       *
 * Version 2 by Brian Calvert.                                               *
 *                                                                           *
 * Software Development Group                                                *
 * National Center for Supercomputing Applications                           *
 * University of Illinois at Urbana-Champaign                                *
 *                                                                           *
 * This is BETA release software.  As such it may contain software bugs and  *
 * exhibit inconsistencies.                                                  *
 *                                                                           *
 * Please send bug reports to polyview@ncsa.uiuc.edu.                        *
 *                                                                           *
 * Copyright (c) 1992 The Board of Trustees of the University of Illinois.   *
 *                                                                           *
 * Permission to use, copy, and modify this software and its                 *
 * documentation for educational, research, and non-profit purposes is       *
 * hereby granted, provided that the above copyright notice, the original    *
 * authors names, and this permission notice appear in all such copies.      *
 * Any distribution of this software requires the explicit and written       *
 * authorization of the authors.                                             *
 *                                                                           *
 * The University of Illinois makes no representations about the             *
 * suitability of this software for any purpose.  It is provided "as is"     *
 * without warranty of any kind.                                             *
 *****************************************************************************/

/* $Id: pvdata.c,v 1.2 93/07/13 16:33:50 gbourhis Exp $ */

#ifdef RCSLOG
$Log:	pvdata.c,v $
 * Revision 1.2  93/07/13  16:33:50  gbourhis
 * fix bugs : test refs <= 1 instead of >= 1, add calls to PVFREE().
 * 
 * Revision 1.1  1992/09/18  10:55:26  marca
 * Initial revision
 *
#endif

#include "pv.h"


/* This references parse_table in pvtext.c. */
extern parse_t parse_table[];

/* How do these fit into new geometry management? */
#define	DEF_WINX 7
/* #define DEF_WINY (getgdesc(GD_YPMAX)-DEF_WINHEIGHT+DEF_WINDY-7) */
#define DEF_WINY (1280-DEF_WINHEIGHT+DEF_WINDY-7)

#define	DEF_WINDX 24
#define	DEF_WINDY -24
#define	DEF_WINWIDTH NTSC_XSIZ
#define	DEF_WINHEIGHT NTSC_YSIZ


/* This is called ONCE, from initialize in pvmain.c. */
int init_data (state_t *state)
{
  /* No windows exist yet, so set all associated window pointers to */
  /* NULL. */
  state->active_windows = NULL;
  state->current_window = NULL;
  state->palette_window = NULL;
  state->status = NULL;
  state->info_window = NULL;
  state->animate_window = NULL;
  state->windows = NULL;
  
  /* Initialize window positions and sizes. */
  state->geometry = AUTO;
  state->nextx = DEF_WINX;
  state->nexty = DEF_WINY;
  state->incx = DEF_WINDX;
  state->incy = DEF_WINDY;
  state->sizex = DEF_WINWIDTH;
  state->sizey = DEF_WINHEIGHT;
  
  /* Initialize the script and transcript file pointers to NULL.  The */
  /* initialization routine will have to override these values if */
  /* files were actually specified on the command line. */
  state->script_fp = NULL;
  state->transcript_fp = NULL;
  
  /* Initialize the parse table to be used for scripts and direct user */
  /* input.  The current flag is set up to add commands */
  /* serially to the end of the list. */
  state->parse_table = parse_table;
  state->current = FALSE;
  
  /* Initialize the action list to show no pending actions. */
  state->action_head = NULL;
  state->action_tail = NULL;
  state->action_free = NULL;
  
  /* No file data or frames have been loaded or created yet, so their */
  /* associated pointers should be NULL. */
  state->files = NULL;
  state->frames = NULL;
  
  /* Initialize new GUI fields. */
  state->base = NULL;
  state->app_context = NULL;
  state->gl_initialized = 0;
  state->exitbox = NULL;
  /* state->help = NULL; */

  /* Assume we don't want anything happening at this point. */
  state->no_actions = TRUE;
  state->last_cmd_already_added = FALSE;
  state->next_window_dtm = 0;

#ifdef DO_AUDIO
  /* Don't use sound unless otherwise informed. */
  state->can_use_sound = 0;
  state->use_sound = 0;
#endif
  
  return ST_OKAY;
}


/* Windows: */
/* This routine only does data structure work also. */
window_t *create_window(state_t *state, char *name)
{
  window_t *new;
  int i;
    
  /* Allocate space for the new window record. */
  new = PVNEW(window_t);
  
  /* Create the root object and increment its reference count by one. */
  WIN_ROOTOBJINFO(new) = create_objinfo(state);
  OIN_REFS(WIN_ROOTOBJINFO(new))++;
  
  /* Initialize the record with suitable defaults. */
  WIN_FOV(new) = DEF_FOV;
  WIN_MAXWIDTH(new) = 0.0;
  WIN_NOTIFY_FN(new) = NULL;
  WIN_SHADE(new) = GOURAUD;

  /* Initially, all objects should be selected. */
  /* NEW: pickmode defaults to faces. */
  WIN_ROOTOBJINFO(new)->pickmode = PICK_FACE;
  
  /* Initialize the file and group names.  Set */
  /* all file pointers to NULL pending opens. */
  strcpy(WIN_GROUPNAME(new), DEF_GROUPNAME);
  strcpy(WIN_INPALETTE(new), "inpal.hdf");
  strcpy(WIN_OUTPALETTE(new), "outpal.hdf");
  strcpy(WIN_RASTERNAME(new), "ris");
  new->rastercont = FALSE;
  new->rasterblastcont = FALSE;
  WIN_SOURCE(new) = NULL;
  WIN_GROUP(new) = NULL;

#ifdef PRIVATE_SUPPORT
  /* By default, windows are not private. */
  new->private = 0;
#endif

  /* Initialize lighting. */
  WIN_LIGHT(new) = 0;
  WIN_LIGHT_TWOSIDE(new) = 1;
  new->legend_flag = 0;
  new->credits_flag = 0;
  new->bounce = 0;

  new->backdrop = PV_PALETTE_BACKG_BLACK;
  new->outline_color = 0x0;
  new->line_width = 1;

  /* NEW: The pickmode defaults to FACE. */
  new->pickmode = PICK_FACE;

  /* NEW: Movepath defaults to spherical. */
  new->movepath = SPHERICAL;

  /* Initialize spheres. */
  new->sphere_size = 0.0;
  new->sphere_depth = 2;
  
  /* Initialize statistics. */
  for (i = 0; i < MAXVDATAS; i++) 
    {
      WIN_STATS(new, i).type = UNKNOWN;
      WIN_STATS(new, i).rec_count = 0;
      WIN_STATS(new, i).rec_size = 0;
      WIN_STATS(new, i).min = FLT_MAX;  /* from limits.h */
      WIN_STATS(new, i).mean = 0.0;
      WIN_STATS(new, i).max = -FLT_MAX;
      WIN_STATS(new, i).mean_sq = 0.0;
      WIN_STATS(new, i).variance = 0.0;
    }
  
  /* Return a pointer to the new record. */
  return new;
}



/* init_window initializes window parameters to reflect changes to related */
/* data structures.  Primarily, it updates statistics related to the window */
/* propagated from the frames of the window.  Requires a pointer to the */
/* system state and the window to be reinitialized.  Returns a pointer to */
/* the window, NULL if an error occurs. */
window_t *init_window (state_t *state, window_t *win, int reset)
{
  frame_t *frame;
  int i;
  
  if ((win != NULL) && (WIN_FRAME(win) != NULL) &&
      (WIN_ROOTOBJ(win) != NULL) /* && reset */) 
    {
      /* Accumulate the statistics from all the child frames of the */
      /* win. */
      for (i = 0; i < MAXVDATAS; i++)
        combine_stats(state, win->stats,
                      win->frame_list->object->stats);
      FOR_EACH_FRAME(frame, i, win)
        combine_stats(state, win->stats,
                      FRA_ROOTOBJ(frame)->stats);
    }
  
  /* Call the window's unique initialization function */
  if (win->init_fn != NULL)
    WIN_INIT(win)(state, win, reset);
  
  /* */
  /* If the window has a palette associated with it, update the */
  /* palette's min and max values. */
  if (WIN_PALETTE(win) != NULL)
    {
      WIN_PALETTE(win)->min = WIN_STATS(win,MVSCALAR).min;
      WIN_PALETTE(win)->max = WIN_STATS(win,MVSCALAR).max;
    }
  
  animate_reset(state, win);
  
  /* Haaaaaaaahahahaahahahaha! */
  if (state->palette_window != NULL)
    set_redraw (state, state->palette_window);
  
  return win;
}


/* This function only handles data structures; therefore, attributes
   are no longer passed in. */
window_t *add_window (state_t *state, char *name, int type, 
                      int (*init_fn)(), int (*redraw_fn)(), 
                      int (*animate_fn)(), 
                      int (*event_fn)(), int (*notify_fn)(), 
                      int (*destroy_fn)())
{
  window_t *new;
  window_t *prev;
  window_t *win;
    
  /* Create the new window record and associated GL window. */
  new = create_window(state, name);
  
  /* Set the window record fields that we know about to either user-. */
  /* specified values or reasonable defaults. */
  strcpy(WIN_NAME(new), name);
  WIN_TYPE(new) = type;
  WIN_IMAGE(new) = NULL;
  WIN_INIT_FN(new) = init_fn;
  WIN_REDRAW_FN(new) = redraw_fn;
  WIN_ANIMATE_FN(new) = animate_fn;
  WIN_EVENT_FN(new) = event_fn;
  WIN_NOTIFY_FN(new) = notify_fn;
  WIN_DESTROY_FN(new) = destroy_fn;
  WIN_PLSREDRAW(new) = TRUE;
  WIN_FRAMES(new) =
    WIN_FRAMEID(new) = 0;
  WIN_FRAMELIST(new) =
    WIN_FRAME(new) = NULL;
  WIN_DELAY(new) = 0;
  WIN_SKIP(new) = 0;
  new->shifted = new->shift_changed = 0;
  
  /* x, y, width, height, cx, and cy are all set by the redraw_windows */
  /* routine before each redraw. */
  
  /* Insert the new record into the list of windows.  First, find the */
  /* record to insert before. */
  for (win = get_window(state, NULL), prev = NULL;
       (win != NULL) && (type > win->type) && 
       (strcasecmp(name, win->name) < 0);
       prev = win, win = win->next)
    ;
  
  /* Then insert the record. */
  if (prev == NULL) {
    /* Special case:  when prev is NULL the window is added at the head */
    /* of the list. */
    new->next = state->windows;
    state->windows = new;
  }
  else {
    /* Otherwise, stick the new window record between prev and win. */
    new->next = prev->next;
    prev->next = new;
  }
  
  /* Initialize the window's data structures. */
  init_window(state, new, FALSE);
  
  return new;
}


window_t *get_window (state_t *state, window_t *win)
{
  if (win == NULL)
    /* If we get a NULL window pointer, return a pointer to the first */
    /* window in the list. */
    return state->windows;
  else 
    /* Return the pointer to the next window in the list. */
    return win->next;
}


window_t *find_window_by_id (state_t *state, long id)
{
  window_t *win;

  /* This routine deals with Xmx-created uniqid's now, not
     GL window ID's.  Methodology should be the same as before. */
  /* Scan through the list of windows until a matching id is found. */
  for (win = get_window(state, NULL); win != NULL;
       win = get_window(state, win))
    if (win->id == id)
      /* Found a match, return pointer to the window. */
      return win;
  
  /* If no match was found, return NULL. */
  return NULL;
}


window_t *find_window_by_name (state_t *state, char *name, int type)
{
  window_t *win;
  
  /* Scan through the list of windows until a matching type and name */
  /* is found. */
  /* Beats me whta that second compare is good for... */
  for (win = get_window(state, NULL); win != NULL;
       win = get_window(state, win))
    if ((win->type == type) /* && (*(win->name) == *name) */ &&
        (strcasecmp(win->name, name) == 0))
      return win;
  
  /* If no match was found, return NULL. */
  return NULL;
}


int delete_window(state_t *state, window_t *win)
{
  window_t *before;
  
  assert (win != NULL);

  if (win == state->windows) 
    {
      /* If the window is the first in the list, it is a special case. */
      state->windows = win->next;
    }
  else 
    {
      /* Let everyone know that the window is closing. */
      broadcast_msg(state, win, MSG_CLOSING);
      
      /* We need to find the window before the one being deleted. */
      /* Search for the window that comes before "window" in the */
      /* list of windows. */
      for (before = get_window(state, NULL);
           (before != NULL) && (before->next != win);
           before = get_window(state, before))
        ;
      
      /* If no window was found, a system error occurred. */
      assert (before != NULL);
      
      /* Otherwise, disconnect the window structure from the list. */
      before->next = win->next;
    }
  
  if (win->destroy_fn != NULL)
    WIN_DESTROY(win)(state, win);

  if (state->active_windows == win)
    state->active_windows = NULL;
  
  PVFREE(win);
  
  return ST_OKAY;
}



/*  Frames: */
frame_t *create_frame(state_t *state, window_t *win, char *name)
{
  frame_t			*frame;
  objinfo_t		*objinfo;
  object_t		*object;
    
  /* Allocate space for the record. */
  frame = PVNEW(frame_t);
  
  /* If there's already a root object info record, get it. */
  /* Otherwise, make a new one. */
  if (WIN_ROOTOBJINFO(win) != NULL)
    objinfo = WIN_ROOTOBJINFO(win);
  else
    /* Make a new object information record. */
    objinfo = create_objinfo(state);
    
  /* Create a root object, set all of its fields active. */
  object = add_object(state, frame, NULL, NULL, OBJECT);
  
  /* Add the appropriate object information record to the root object. */
  add_objinfo(state, object, objinfo);
  
  /* Initialize the fields of the frame. */
  strcpy(FRA_NAME(frame), name);
  FRA_ROOTOBJ(frame) = object;
  FRA_NEXT(frame) = FRA_PREV(frame) = NULL;
  
  return frame;
}


int build_matching_frames(state_t *state, window_t *win, char *path,
                          char *location, Vgroup_t *group)
{
  char *p;
  char localname[MAXNAMELEN];
  int matches;
  Vgroup_t *vg;
  frame_t *frame;
    
  /* Quit if the group is bad. */
  if (group == NULL) 
    {
      fprintf(stderr, "[build_matching_frames] ERROR: Bad group.\n");
      fflush (stderr);
      return 0;
    }
  
  /* Make sure that all the children of this vgroup are in memory. */
  if (VGR_CHILD_IN(group) != TRUE)
    build_vgroup_children(state, group);
  
  if (*location == '/') 
    {
      /* Special case if leading root group slash. */
      location++;
      localname[0] = '/';
      localname[1] = '\0';
      vg = group;
    }
  else 
    {
      /* Copy the next substring */
      for (p = localname; (*location != '\0') && (*location != '/');
           location++, p++)
        *p = *location;
      *p = '\0';
      vg = VGR_GROUPS(group); 
      
      if (*location != '\0')
        location++;		/* skip slash */
    }
  
  /* If this is the end of the path, find all the matching vdatas */
  matches = 0;
  for ( ; vg != NULL; vg = VGR_SIBLING(vg)) 
    {
      if (strwcmp(localname, VGR_NAME(vg)) == 0) 
        {
          if (*location == '\0') 
            {
              /* Reached the end of the path.  Add frame */
              /* corresponding to group. */
              frame = add_frame(state, win, VGR_NAME(vg));
              if (build_frames_r(state, win, FRA_ROOTOBJ(frame), vg) != 
                  ST_OKAY) 
                {
                  fprintf (stderr, "[build_matching_frames] ERROR: %s\n",
                           "Should delete all of the frames at this point.");
                  fflush (stderr);
                  /* Do we return here? */
                }
              
              combine_stats(state, win->stats,
                            FRA_ROOTOBJ(frame)->stats);
              matches++;
            }
          else 
            {
              matches = 
                build_matching_frames(state, win, path, location, vg);
            }
        }
    }
  
  return matches;
}


frame_t *add_frame(state_t *state, window_t *win, char *name)
{
  frame_t *frame;
  frame_t *first;
  frame_t *last;
    
  /* Create the new frame. */
  frame = create_frame(state, win, name);
  
  /* Add the frame to the window's frame list. */
  first = WIN_FRAMELIST(win);
  if (first == NULL) 
    {
      /* First frame of window. */
      OBJ_INFO(FRA_ROOTOBJ(frame)) = WIN_ROOTOBJINFO(win);
      
      WIN_FRAMELIST(win) = WIN_FRAME(win) = frame;
      
      FRA_NEXT(frame) = FRA_PREV(frame) = frame;
    }
  else 
    {
      /* Not the first frame.  Add to the end of the list. */
      /* Remember, this is doubly linked and circular. */
      OBJ_INFO(FRA_ROOTOBJ(frame)) = WIN_ROOTOBJINFO(win);
      last = FRA_PREV(first);
      
      FRA_PREV(first) = frame;
      FRA_NEXT(last) = frame;
      
      FRA_PREV(frame) = last;
      FRA_NEXT(frame) = first;
    }
  
  /* Up the count of frames. */
  WIN_FRAMES(win)++;
  
  return frame;
}


/* build frames creates the data structures corresponding to those found */
/* in the vgroup structure passed to it.  The function requires a state */
/* record pointer, pointer to the window which will hold the frames, */
/* and a pointer to the vgroup data structure in memory that will be used */
/* as a source of data.  The recursive function to add subobjects is called */
/* to handle the rest of the structure.  Returns ST_OKAY if no errors occur. */
int build_frames(state_t *state, window_t *win, Vgroup_t *group)
{
  frame_t *prev;
  frame_t *curr;
  int status, i;

  if (group == NULL)
    {
      stprintf (state, "ERROR:  No data to load.\n");
      XmxMakeErrorDialog (win->base,
                          "ERROR: No data to load.",
                          "Error While Loading Data");
      XtManageChild (Xmx_w);
      return ST_ERROR;
    }
  
  /* Build the frames using specified information.  If no matching */
  /* frames exist, abort the building process. */
  if (build_matching_frames
      (state, win, WIN_GROUPNAME(win),
       WIN_GROUPNAME(win), WIN_SOURCE(win)->groups) == 0)
    {
      stprintf(state, "ERROR:  No frames match '%s'.\n", WIN_GROUPNAME(win));
      XmxMakeErrorDialog (win->base,
                          "ERROR: Cannot find any matching frames.",
                          "Error While Loading Data");
      XtManageChild (Xmx_w);
      return ST_ERROR;
    }
  
  /* Step through the object trees associated with each frame. */
  /* Fill in derived data sets for all of the objects.  Derived data */
  /* includes COORD and COLOR data sets.  If an object is still */
  /* missing objects or vdatas which exist in the previous frame, the */
  /* missing items are copied from the previous frame. */

  /* This calls consistent_data, which may open a dialog box. */
  status = 
    fill_in_frame(state, win, WIN_FRAMELIST(win), WIN_FRAMELIST(win));
  for (prev = WIN_FRAMELIST(win), curr = WIN_FRAMELIST(win)->next;
       (curr != WIN_FRAMELIST(win)) && (status == ST_OKAY);
       prev = curr, curr = curr->next) 
    status = fill_in_frame(state, win, prev, curr);
  
  if (status == ST_OKAY) 
    {
      /* Initialize the frame id. */
      if (WIN_FRAMELIST(win) != NULL)
        WIN_FRAMEID(win) = 1;
    }
  else 
    {
      /* If an error was detected while filling the frames, free */
      /* all of them and quit the effort. */
      stprintf (state, "ERROR:  Could not load frames.");
      /* Don't open a dialog box; it looks like consistent_data will
         have taken care of this for us. */
      data_clear(state, win);
    }
  
  return status;
}


/* A recursive function called to create drawing objects from the vgroup */
/* hierarchy.  User-defined vdata names are used to identify the vdatas */
/* to be used for each object data set (connectivity, scalar values, */
/* and coordinates). */
int build_frames_r (state_t *state, window_t *win, object_t *object, 
                    Vgroup_t *group)
{
  object_t *newobj;
  Vdata_t *vdata;
  
  /* Make sure that all the children of this vgroup are in memory. */
  if (VGR_CHILD_IN(group) != TRUE)
    build_vgroup_children(state, group);
  
  /* If this group is empty... */
  if (VGR_GROUPS(group) == NULL) 
    {
      /* This group is the only one in this branch, so try to read */
      /* vdatas to fill it up. */
      OBJ_TYPE(object) = ATOM;
      
      /* Get vdatas with the proper names and assign them to the */
      /* object for separate or common x, y, and z coordinates. */
      vdata = get_vdata_in_vgroup(state, group, WIN_PXNAME(win));
      set_object_data(state, object, PX, vdata);
      
      vdata = get_vdata_in_vgroup(state, group, WIN_PYNAME(win));
      set_object_data(state, object, PY, vdata);
      
      vdata = get_vdata_in_vgroup(state, group, WIN_PZNAME(win));
      set_object_data(state, object, PZ, vdata);

      /* Try to get aggregate coord. */
      vdata = get_vdata_in_vgroup(state, group, win->vname[PCOORD]);
      set_object_data(state, object, PCOORD, vdata);
      
      /* Get whatever connectivity vdata is available and */
      /* assign it to the object. */
      vdata = get_vdata_in_vgroup(state, group, WIN_PLISTNAME(win));
      set_object_data(state, object, MCONNECT, vdata);
      
      /* Assign a "color" data set to the object. */
      if ((vdata = get_vdata_in_vgroup(state, group, "color")) == NULL) 
        {
          /* No color vdata exists.  Assign the scalar values */
          /* (if any) to the object. */
          vdata = get_vdata_in_vgroup(state, group,
                                      WIN_VSCALARNAME(win));
          set_object_data(state, object, MVSCALAR, vdata);
          set_object_data(state, object, VCOLOR, NULL);
        }
      else 
        {
          /* There is a color vdata already.  Use it for the */
          /* object. */
          set_object_data(state, object, VCOLOR, vdata);
          set_object_data(state, object, MVSCALAR, NULL);
        }
    }
  else 
    {
      /* Assume that the vgroup is just a group since it has no */
      /* children that are vdatas.  Recursively build the children. */
      for (group = VGR_GROUPS(group); group != NULL;
           group = VGR_SIBLING(group)) 
        {
          
          /* Create a new object and initialize its name. */
          newobj = add_object(state, NULL, object, NULL, OBJECT);
          if (newobj == NULL) 
            {
              stprintf(state, "ERROR:  Too many objects in Vgroup.\n");
              return ST_ERROR;
            }
          OBJ_NAME(newobj) = VGR_NAME(group);
          
          /* Recursively build the objects under the frame. */
          if (build_frames_r(state, win, newobj, group) != ST_OKAY)
            return ST_ERROR;
          
          /* Propagate the statistics from the new child */
          /* object to the parent. */
          combine_stats(state, object->stats, newobj->stats);
        }
    }
  
  return ST_OKAY;
}


int fill_in_frame (state_t *state, window_t *win, frame_t *source, 
                   frame_t *dest)
{
  /* Check that the pointers are legal. */
  assert (source != NULL);
  assert (dest != NULL);
  
  /* If there is no root object for the source frame, there is no point */
  /* in copying. */
  if (FRA_ROOTOBJ(source) == NULL)
    return ST_OKAY;
  
  /* Recursively fill in the objects under this tree. */
  if (fill_in_objects(state, win, FRA_ROOTOBJ(source), FRA_ROOTOBJ(dest)) != ST_OKAY)
    return ST_ERROR;

  return ST_OKAY;
}


int fill_in_objects(state_t *state, window_t *win, object_t *source,
                    object_t *dest)
{
  int type;
  object_t *parent;
  
  /* Make sure that the objects are of the same type.*/
  if (OBJ_TYPE(source) != OBJ_TYPE(dest))
    return ST_ERROR;

  type = OBJ_TYPE(source);
  
  /* If the source does not have an objinfo record, create one and */
  /* add it. */
  if (OBJ_INFO(source) == NULL)
    add_objinfo(state, source, NULL);
  
  /* If the dest does not have an objinfo record, link it to the one */
  /* that the source uses. */
  if (OBJ_INFO(dest) == NULL)
    OBJ_INFO(dest) = OBJ_INFO(source);
  
  /* If the object is an ATOM, copy the vdata slots that are missing. */
  if (type == ATOM) 
    {
      /* Derive all of the data possible. */
      derive_data(state, win, source, dest);
      
      /* Make sure that all of the object's vdata slots are filled. */
      fill_in_vdatas(state, source, dest, 0, MAXVDATAS-1);
      
      /* Check the data assigned to the object for consistency. */
      PVD (("** Calling consistent_data on obj '%s'\n", dest->name));
      if (consistent_data(state, win, dest) != ST_OKAY) 
        {
          stprintf(state,
                   "ERROR:  Inconsistent data found while loading.\n");
          return ST_ERROR;
        }
    }
  else if (type == OBJECT) 
    {
      source = source->u.grp.children;
      parent = dest;
      dest = dest->u.grp.children;
      while ((source != NULL) && (dest != NULL)) 
        {
          fill_in_objects(state, win, source, dest);
          combine_stats(state, parent->stats, dest->stats);
          source = source->sibling;
          dest = dest->sibling;
        }
      dest = parent;
    }
  
  /* In either case, set the initial drawing function. */
  if (dest->info != NULL)
    dest->info->draw_fn = WIN_DRAW(win);
  
  return ST_OKAY;
}

int fill_in_vdatas(state_t *state, object_t *source, object_t *dest,
                   int low, int high)
{
  int i;
  
  while (low <= high) 
    {
      if (dest->u.obj.vdata[low] == NULL) 
        {
          dest->u.obj.vdata[low] = source->u.obj.vdata[low];
          for (i = 0; i < MAXDATADIMS; i++)
            dest->stats[i][low] = source->stats[i][low];
        }
      
      low++;
    }
  return ST_OKAY;
}


float *merge_triple(int num, float *px, float *py, float *pz)
{
  float			*triple;
  float			*t;
  
  /* Make space for the triple. */
  t = triple = PVCALLOC(3 * num, float);
  
  /* Copy the three arrays into the triple. */
  while (num > 0) 
    {
      if (px == NULL)
        *(t++) = 0.0;
      else
        *(t++) = *(px++);
      
      if (py == NULL)
        *(t++) = 0.0;
      else
        *(t++) = *(py++);
      
      if (pz == NULL)
        *(t++) = 0.0;
      else
        *(t++) = *(pz++);
      
      num--;
    }
  
  /* Return the triple.  The caller is responsible for freeing it. */
  return triple;
}


char *get_data(state_t *state, Vdata_t *vdata)
{
  if (vdata == NULL)
    return NULL;
  else 
    {
      vdata = get_vdata(state, vdata);
      return vdata->data;
    }
}


int consistent_data(state_t *state, window_t *win, object_t *obj)
{
  int i, size, nvert;
  stats_t *stats;
  Vdata_t *vdata;
  
  /* This function performs a number of checks to insure that data */
  /* loaded into the current object is self-consistent... */

  PVD (("[entering cons] obj is '%s'\n", obj->name))
  
  /* CHECK 1:  All objects must have a set of coordinates, no matter */
  /* how small. */
  vdata = OBJ_VDATA(obj, PCOORD);
  if (vdata == NULL) 
    {
      stprintf(state, 
               "ERROR: Object %s must have coordinate information.",
               OBJ_NAME(obj));
      XmxMakeErrorDialog (win->base, 
                          "ERROR: No coordinate information in this Vset.",
                          "Data Consistency Error");
      XtManageChild (Xmx_w);
      return ST_ERROR;
    }
  
  /* CHECK 2:  The coordinate list must be a set of floating point */
  /* triplets. */
  stats = &OBJ_STATS(obj, PCOORD, 0);
  if ((stats->type != PVFLOAT) && (stats->rec_size != 3)) 
    {
      stprintf(state,
               "ERROR: Object %s (coordinates) must be float triplets.", 
               OBJ_NAME(obj));
      XmxMakeErrorDialog (win->base, 
                          "ERROR: Coordinate data are not float triplets.",
                          "Data Consistency Error");
      XtManageChild (Xmx_w);
      return ST_ERROR;
    }
  
  /* This is needed by other checks. */
  nvert = stats->rec_count;
  PVD (("[cons] nvert = %d\n", nvert));
  
  /* If there is a connectivity list... */
  if (OBJ_VDATA(obj, MCONNECT) != NULL) 
    {
      PVD (("[cons] MCONNECT vdata is %s\n", OBJ_VDATA(obj, MCONNECT)->name));

      /* CHECK 3:  The connectivity list must be a set of integer */
      /* vectors. */
      stats = &OBJ_STATS(obj, MCONNECT, 0);
      
      if ((stats->type != PVINTEGER) || (stats->rec_size <= 1)) 
        {
          stprintf
            (state,
             "ERROR: Object %s (connectivity) must be integer vectors.",
             OBJ_NAME(obj));
          XmxMakeErrorDialog 
            (win->base, 
             "ERROR: Connectivity data must be integer vectors.",
             "Data Consistency Error");
          XtManageChild (Xmx_w);
          return ST_ERROR;
        }
      
      /* CHECK 4:  Each element of the connectivity list must */
      /* reference no vertex id out of the range of ids (0 through */
      /* n, where n is the number of vertices). */
      /* WHAT A BONEHEAD... */
      /* for (i = stats->rec_size; i > 0; i--) */
      for (i = 0; i < stats->rec_size; i++)
        {
          PVD (("  [cons] col %d min is %f max is %f\n", i, OBJ_STATS(obj, MCONNECT, i).min, OBJ_STATS(obj, MCONNECT, i).max));

          /* Check that no value is less than zero. */
          if ((int)OBJ_STATS(obj, MCONNECT, i).min < 0) 
            {
              stprintf
                (state, 
                 "ERROR: Object %s (connectivity) contains negative values.",
                 OBJ_NAME(obj));
              XmxMakeErrorDialog 
                (win->base, 
                 "ERROR: Connectivity contains negative values.",
                 "Data Consistency Error");
              XtManageChild (Xmx_w);
              return ST_ERROR;
            }
          
          /* Check that no value is greater than the number of */
          /* vertices. */
          if ((int)OBJ_STATS(obj, MCONNECT, i).max > nvert) 
            {
              stprintf
                (state, 
                 "ERROR: Object %s (connectivity) contains excessively high values.",
                 OBJ_NAME(obj));
              fprintf (stderr, "max %d nvert %d\n", 
                       (int)OBJ_STATS(obj, MCONNECT, i).max, nvert);
              XmxMakeErrorDialog (win->base, 
                                  "ERROR: Connectivity values are too high.",
                                  "Data Consistency Error");
              XtManageChild (Xmx_w);
              return ST_ERROR;
            }
        }
    }
  
  /* If there is a color list... */
  if (OBJ_VDATA(obj, VCOLOR) != NULL) 
    {
      /* CHECK 5:  The color list must be a set of integer scalars. */
      stats = &OBJ_STATS(obj, VCOLOR, 0);

#if 0
      PVD (("[consistent_data], doing color vdata:\n"));
      PVD (("  vdata->name is %s\n", OBJ_VDATA(obj, VCOLOR)->name));
      PVD (("  vdata->id is %d\n", OBJ_VDATA(obj, VCOLOR)->id));
      PVD (("  stats->type is %d\n", stats->type));
      PVD (("  stats->rec_size is %d\n", stats->rec_size));
#endif

      if ((stats->type != PVINTEGER) || (stats->rec_size > 1)) 
        {
          stprintf(state, "ERROR: Object %s (color) must be integer scalars.",
                   OBJ_NAME(obj));
          return ST_ERROR;
        }
      
      /* CHECK 6:  The color list must be of length n, where n is */
      /* the number of vertices. */
      if (stats->rec_count < nvert)
        {
          stprintf(state, 
                   "ERROR: Object %s (color) must have as many values as points exist.",
                   OBJ_NAME(obj));
          return ST_ERROR;
        }
      
      /* CHECK 7:  All color values must be greater than zero and */
      /* span a range of no more than SPEC_SIZE colormap entries. */
      /* Check that no value is less than zero. */
      if ((int)stats->min < 0) 
        {
          stprintf(state, "ERROR: Object %s (color) contains negative values.",
                 OBJ_NAME(obj));
          return ST_ERROR;
        }
      
      /* Check that the range spanned is less than SPEC_SIZE */
      /* entries. */
      if (((int)stats->max - (int)stats->min) >= SPEC_SIZE) 
        {
          stprintf(state, 
                   "ERROR: Object %s color vdata contains excessively large values.",
                   OBJ_NAME(obj));
          return ST_ERROR;
        }
    }
  
  return ST_OKAY;
}


int derive_data (state_t *state, window_t *win, object_t *prev, object_t *obj)
{
  Vdata_t *vscalar, *vcolor;
  Vdata_t *vdata;
  Vdata_t **vdatas;
  float	*px, *py, *pz;
  int nvert;
  
  /* Initialize a pointer to the object's list of vdatas. */
  vdatas = obj->u.obj.vdata;
  
  /* Get the sets of x, y, and z-coordinates, either from this */
  /* frame or from the previous one.  Only copy pcoords if there */
  /* are no px, py, and pz data sets. */
  if ((OBJ_VDATA(prev,PX) == NULL) && (OBJ_VDATA(prev,PY) == NULL) &&
      (OBJ_VDATA(prev,PZ) == NULL)) {
    fill_in_vdatas(state, prev, obj, PCOORD, PCOORD);
  }
  else {
    fill_in_vdatas(state, prev, obj, PX, PZ);
  }
  
  /* If there is no pcoord data set, create one.  Otherwise, copy the */
  /* pcoord statistics from each part of the triple to the individual */
  /* data sets. */
  if ((vdata = get_vdata(state, vdatas[PCOORD])) == NULL) {
    
    /* If there are no coordinate data sets, warn the user */
    /* and quit this effort. */
    px = (float *) get_data(state, vdatas[PX]);
    py = (float *) get_data(state, vdatas[PY]);
    pz = (float *) get_data(state, vdatas[PZ]);
    if ((px == NULL) && (py == NULL) && (pz == NULL)) 
      {
        stprintf(state,
                 "WARNING:  No matching x, y, or z vdatas");
        return ST_ERROR;
      }
    
    /* Create the PCOORD vdata. */
    OBJ_VDATA(obj,PCOORD) = 
      vdata = create_vdata(state, -1, "pcoord", NULL);
    nvert = MAX(MAX(OBJ_STATS(obj, PX, 0).rec_count,
                    OBJ_STATS(obj, PY, 0).rec_count),
                OBJ_STATS(obj, PZ, 0).rec_count);
    VDA_DATA_IN(vdata) = TRUE;
    vdata->data = (char *) merge_triple(nvert, px, py, pz);
    
    VDA_STATS(vdata).type = PVFLOAT;
    VDA_STATS(vdata).rec_count = nvert;
    
    /* Only copy statistical information if the corresponding */
    /* data set exists. */
    if (px != NULL) 
      {
        OBJ_STATS(obj, PCOORD, X) =
          vdata->stats[X] = VDA_STATS(vdatas[PX]);
      }
    if (py != NULL) 
      {
        OBJ_STATS(obj, PCOORD, Y) =
          vdata->stats[Y] = VDA_STATS(vdatas[PY]);
      }
    if (pz != NULL) 
      {
        OBJ_STATS(obj, PCOORD, Z) =
          vdata->stats[Z] = VDA_STATS(vdatas[PZ]);
      }
  }
  else 
    {
      OBJ_STATS(obj,PX,0) = OBJ_STATS(obj,PCOORD,X);
      OBJ_STATS(obj,PY,0) = OBJ_STATS(obj,PCOORD,Y);
      OBJ_STATS(obj,PZ,0) = OBJ_STATS(obj,PCOORD,Z);
    }
  
  /* If there is a vscalar data set and no vcolor data set already */
  /* build one. */
  vcolor = get_vdata(state, vdatas[VCOLOR]);
  vscalar = get_vdata(state, vdatas[MVSCALAR]);
  if ((vcolor == NULL) && (vscalar != NULL)) 
    {
      float *scalar;
      int *idata;
      int icolor;
      float min, max;
      int i, n;
      
      scalar = VDA_DATA(vscalar,float);
      min = WIN_STATS(win, MVSCALAR).min;
      max = WIN_STATS(win, MVSCALAR).max;
      nvert = VDA_STATS(vscalar).rec_count;
      
      /* Create the VCOLOR vdata. */
      OBJ_VDATA(obj,VCOLOR) = 
        vdata = create_vdata(state, -1, "vcolor", NULL);
      VDA_DATA_IN(vdata) = TRUE;
      vdata->data = (char *) PVMALLOC(nvert*sizeof(int));
      idata = VDA_DATA(vdata, int);
      
      VDA_STATS(vdata).type = PVINTEGER;
      VDA_STATS(vdata).rec_count = nvert;
      /* New fix from Brian Calvert -Marc */
      VDA_STATS(vdata).min = SPEC_BASE+1;
      VDA_STATS(vdata).max = SPEC_BASE+SPEC_SIZE-2;
      /* End of new fix. */
      
      OBJ_STATS(obj,VCOLOR,0) = VDA_STATS(vdata);
      
      for (i = 0, n = nvert; i < n; i++) 
        {
          icolor = MAP_COLOR(min, *(scalar++), max);
          
          if (icolor <= SPEC_BASE) 
            icolor = SPEC_BASE+1;
          else if (icolor >= (SPEC_BASE+SPEC_SIZE-1)) 
            icolor = SPEC_BASE+SPEC_SIZE-2;
          
          (*idata++) = icolor;
        }
    }
  
  return ST_OKAY;
}


frame_t *get_next_frame(state_t *state, frame_t *frame)
{
  if (frame == NULL)
    return NULL;
  else
    return FRA_NEXT(frame);
}


frame_t *get_prev_frame(state_t *state, frame_t *frame)
{
  if (frame == NULL)
    return NULL;
  else
    return FRA_PREV(frame);
}


frame_t *get_first_frame(state_t *state, window_t *win)
{
  return WIN_FRAMELIST(win);
}


frame_t *find_frame_by_name(state_t *state, window_t *win, char *name)
{
  frame_t *frame;
  
  /* Begin the search at the top of the list of frames. */
  frame = WIN_FRAMELIST(win);
  
  /* Search the list from the first frame to the last, looking for */
  /* matches to the name.  The name may be wildcarded, but only the */
  /* first instance is returned. */
  while (frame != NULL) 
    {
      /* Is this the frame?  If so, return a pointer to it. */
      if (strwcmp(name, FRA_NAME(frame)) == 0)
        return frame;
      
      /* Get the next frame and make sure that we have not wrapped */
      /* around to the beginning of the list. */
      frame = get_next_frame(state, frame);
      if (frame == WIN_FRAMELIST(win)) 
        {
          /* Search has completed, nothing found. */
          return NULL;
        }
    }
  
  return NULL;
}


frame_t *find_frame_by_num(state_t *state, window_t *win, int num)
{
  frame_t *frame;
  int frame_number;
  
  /* If the frame number is larger than the number of frames, don't */
  /* bother. */
  if (num > WIN_FRAMES(win))
    return NULL;
  
  /* Begin the search at the top of the list of frames. */
  frame = WIN_FRAMELIST(win);
  frame_number = 1;
  
  /* Search the list from the first frame to the last, looking for */
  /* matches to the name.  The name may be wildcarded, but only the */
  /* first instance is returned. */
  while (frame != NULL) 
    {
      /* Is this the frame?  If so, return a pointer to it. */
      if (frame_number == num)
        return frame;
      
      /* Get the next frame and make sure that we have not wrapped */
      /* around to the beginning of the list. */
      frame = get_next_frame(state, frame);
      if (frame == WIN_FRAMELIST(win)) 
        {
          /* Search has completed, nothing found. */
          return NULL;
        }
      
      /* Increment the frame count. */
      frame_number++;
    }
  
  return NULL;
}


int find_frame_num(state_t *state, window_t *win, frame_t *frame)
{
  int framenum;
  frame_t *f;
  
  for (framenum = 1, f = WIN_FRAMELIST(win);
       (f != NULL) && (f != frame);
       framenum++, f = FRA_NEXT(f))
    ;
  
  if (f == NULL)
    return 0;
  else
    return framenum;
}


frame_t *delete_frame(state_t *state, window_t *win, frame_t *frame)
{
  frame_t *next;
  frame_t *prev;
  
  /* If the frame is NULL, no point in deleting it. */
  if (frame == NULL)
    return NULL;
  
  /* Get a pointer to the next frame in the list.  This is the value */
  /* returned. */
  next = get_next_frame(state, frame);
  prev = get_prev_frame(state, frame);
  
  if ( (frame == next) || (frame == prev) ) 
    {
      /* This is apparently the only frame in the list, so the */
      /* next and prev frames should be NULL. */
      next = prev = NULL;
    }
  else 
    {
      /* Unlink the frame from the list. */
      FRA_PREV(next) = prev;
      FRA_NEXT(prev) = next;
    }
  
  /* Free the associated object and the frame. */
  delete_object(state, FRA_ROOTOBJ(frame), TRUE);

  
  /* If the frame was also the first frame in the list, update the */
  /* win's frame list pointer. */
  if (frame == WIN_FRAMELIST(win))
    WIN_FRAMELIST(win) = next;
  
  /* If the frame was also the current frame, then update the window's */
  /* frame pointer. */
  if (frame == WIN_FRAME(win))
    WIN_FRAME(win) = next;
  
  PVFREE(frame);

  WIN_FRAMES(win)--;
  
  return next;
}


/* Objects: */
object_t *create_object(state_t *state, int type)
{
  object_t *object;
  
  /* Create the object. */
  object = PVNEW(object_t);
  
  /* Initialize its fields. */
  OBJ_TYPE(object) = type;
  
  return object;
}


object_t *add_object(state_t *state, frame_t *frame, object_t *parent, 
                     object_t *object, int type)
{
  /* If the parent group is full, abort. */
  if ((parent != NULL) && (OBJ_NUMCHILD(parent) >= MAXOBJECTS)) 
    {
      bprintf(state,
              "ERROR:  Maximum number of subobjects is %d.\n",
              MAXOBJECTS);
      return NULL;
    }
  
  if ((parent == NULL) || (OBJ_TYPE(parent) == OBJECT)) 
    {
      /* Create the new object. */
      object = create_object(state, type);
      
      /* If the parent object is NULL, the parent is really the */
      /* root object of the frame. */
      if (parent == NULL) {
        /* Insert the object as the root object. */
        FRA_ROOTOBJ(frame) = object;
      }
      else {
        /* Insert the object as a child of the parent */
        OBJ_SIBLING(object) = OBJ_CHILD(parent);
        OBJ_CHILD(parent) = object;
        OBJ_NUMCHILD(parent)++;
      }
      OBJ_REFS(object)++;
      
      OBJ_NAME(object) = "<null>";
    }
  else {
    /* The parent object must be a group in order to insert */
    /* another object. */
    bprintf(state,
            "SYSTEM ERROR:  Parent object must be a group.\n");
    object = NULL;
  }
  
  return object;
}


int set_object_data(state_t *state, object_t *object, int vdata_slot,
                    Vdata_t *vdata)
{
  int i;
  
  if (object == NULL) 
    {
      printf("ERROR:  Attempt to set NULL object.\n");
      return ST_ERROR;
    }
  
  if (vdata_slot < MAXVDATAS) 
    {
      OBJ_VDATA(object, vdata_slot) = vdata;
      if (vdata != NULL) 
        {
          swap_in(state, vdata);
          for (i = 0; i < MAXDATADIMS; i++)
            OBJ_STATS(object,vdata_slot,i) =
              vdata->stats[i];
        }
    }
  else 
    {
      bprintf(state, 
              "SYSTEM ERROR:  vdata_slot too large.\n");
      return ST_ERROR;
    }
  
  return ST_OKAY;
}


long ApplyFuncToTree(state_t *state, object_t *tree, long (*fn)(), ...)
{
  va_list args;
  int value;
  
  /* Initialize args to point to the beginning of the list of */
  /* arguments.  This pointer is used for all further recursive */
  /* calls. */
  va_start(args, fn);
  value =  ApplyFuncToTree_r(state, tree, fn, args);
  va_end(args);
  
  return value;
}


long ApplyFuncToTree_r(state_t *state, object_t *object, long (*fn)(),
                       va_list args)
{
  object_t *curr;
  long sum;
  
  sum = 0;
  if (OBJ_TYPE(object) == OBJECT) {
    sum += (*fn)(state, object, args);
    
    /* Call this function on all of the children of this level, */
    /* summing the return value of the function each time. */
    for (curr = OBJ_CHILD(object); curr != NULL;
         curr = OBJ_SIBLING(curr)) {
      sum += ApplyFuncToTree_r(state, curr, fn, args);
    }
  }
  else {
    /* Apply the function to all of the objects on this level, */
    /* summing the return value of the function each time. */
    for (curr = object; curr != NULL; curr = OBJ_SIBLING(curr)) {
      /* Apply the function to the object. */
      sum += (*fn)(state, curr, args);
    }
  }
  
  /* Return the value. */
  return sum;
}



int delete_object(state_t *state, object_t *object, int recursion)
{
  object_t *child;
  
  if (recursion) 
    {
      if (OBJ_TYPE(object) == OBJECT) {
        for (child = OBJ_CHILD(object); child != NULL;
             child = OBJ_SIBLING(child)) {
          /* Only recurse if the object is actually */
          /* being deleted.  That's the only time */
          /* child objects need to be traversed. */
          if (OBJ_REFS(object) > 1)
            recursion = FALSE;
          else
            recursion = TRUE;
          
          delete_object(state, child, recursion);
        }
      }
      else {
        /* Nothing to recurse on. */
        /* FUTURE EXPANSION:  Free reference to vdata. */
        /* vdata would have usage_count.  Could be used as */
        /* first stage of memory free-up, second based on */
        /* last reference time. */
      }
    }
  
  if (OBJ_REFS(object) <= 1) {
    /* No one else is interested.  Free up the space associated */
    /* with this record. */
    delete_objinfo(state, OBJ_INFO(object));
#if 1
    PVFREE(object);
#endif
  }
  else {
    /* There are other users of this object.  Keep it. */
    OBJ_REFS(object)--;
  }
  
  return ST_OKAY;
}




/* Object information: */
/* create objinfo creates a new object information record, initializes it */
/* to cause no changes, and returns a pointer to it.  It only requires a */
/* pointer to the state record.  Returns NULL if an error occurs. */
objinfo_t *create_objinfo(state_t *state)
{
  objinfo_t *info;
    
  /* Create the information record. */
  info = PVNEW(objinfo_t);

  OIN_REFS(info) = 0;
  
  /* Set up picking constants. */
  /* NEW: pickmode defaults to faces */
  info->pickmode = PICK_FACE;
  
  /* Set the scale values to defaults (1, 1, 1). */
  XFO_COORD(OIN_SCL(info), X) = 1.0;
  XFO_COORD(OIN_SCL(info), Y) = 1.0;
  XFO_COORD(OIN_SCL(info), Z) = 1.0;
  
  return info;
}


/* add objinfo adds an object information record to an object.  The function */
/* accepts the system state, object pointer, and object information pointer. */
/* If the object information pointer is NULL, a new record is created, */
/* initalized, and used.  If the object information pointer is not NULL, */
/* its usage count is updated.  Returns a pointer to the object information */
/* record, NULL if an error occurs. */
objinfo_t *add_objinfo(state_t *state, object_t *object, objinfo_t *info)
{
  assert (object != NULL);
  
  /* Check if the objinfo record needs to be created or not. */
  if (info == NULL)
    info = create_objinfo(state);

  OIN_REFS(info)++;
  
  /* Add the object information to the object. */
  OBJ_INFO(object) = info;
  
  return info;
}


/* delete objinfo deletes an unused object information record.  It requires */
/* pointers to the system state and the object information record to be */
/* deleted.  If the object information record has a usage greater than 1, */
/* its value is decremented.  Otherwise, the record is freed. */
int delete_objinfo(state_t *state, objinfo_t *info)
{
  assert (info != NULL);
  
  if (OIN_REFS(info) <= 1)
    /* No one else is interested.  Free up the space associated */
    /* with this record. */
    {
#if 1
      /* THIS causes coredumps. */
      PVFREE(info);
#endif
    }
  else
    /* There are other users of this object info record. */
    OIN_REFS(info)--;
  
  return ST_OKAY;
}


/* Palettes: */

/* create palette makes a new palette record and initializes it to black. */
/* The function expects a pointer to the system state.  Returns pointer to */
/* the new palette record, NULL if an error occurs. */
palette_t *create_palette (state_t *state)
{
  palette_t *new;
  
  /* Allocate space for the new window record. */
  new = PVNEW(palette_t);
  
  /* Initialize the slope and origin of the palette mapping function. */
  new->slope = 1.0;
  new->origin = 0.0;
  
  /* The base of the palette is important so that it doesn't overwrite */
  /* the system palette. */
  new->base = SPEC_BASE;
  
  /* Make sure there is no initial probe value. */
  new->probe_value = -1;
  
  /* Return a pointer to the new record. */
  return new;
}


/* add palette creates a palette, initializes it to the default palette, */
/* and sets the window to point to the palette.  Function requires a */
/* pointer to the system state, window record, and palette.  If the palette */
/* pointer is provided it is used without modification.  Returns a pointer */
/* to the new record, NULL otherwise. */
palette_t *add_palette (state_t *state, window_t *win, palette_t *palette)
{
  /* If no palette was passed in, create one and initialize it */
  /* appropriately. */
  if (palette == NULL) 
    {
      palette = create_palette(state);
      set_palette(state, palette, PAL_DEFAULT);
    }
  
  /* Increment the number of references that are being made to this */
  /* palette. */
  (palette->refs)++;
  WIN_PALETTE(win) = palette;
  
  return palette;
}
