/*****************************************************************************
 * NCSA Polyview 3.1                                                         *
 *                                                                           *
 * Version 3.1 changes and additions by Gilles Bourhis                       *
 * Version 3 changes and additions by Marc Andreessen.                       *
 * Version 2 by Brian Calvert.                                               *
 *                                                                           *
 * Software Development Group                                                *
 * National Center for Supercomputing Applications                           *
 * University of Illinois at Urbana-Champaign                                *
 *                                                                           *
 * This is BETA release software.  As such it may contain software bugs and  *
 * exhibit inconsistencies.                                                  *
 *                                                                           *
 * Please send bug reports to polyview@ncsa.uiuc.edu.                        *
 *                                                                           *
 * Copyright (c) 1992 The Board of Trustees of the University of Illinois.   *
 *                                                                           *
 * Permission to use, copy, and modify this software and its                 *
 * documentation for educational, research, and non-profit purposes is       *
 * hereby granted, provided that the above copyright notice, the original    *
 * authors names, and this permission notice appear in all such copies.      *
 * Any distribution of this software requires the explicit and written       *
 * authorization of the authors.                                             *
 *                                                                           *
 * The University of Illinois makes no representations about the             *
 * suitability of this software for any purpose.  It is provided "as is"     *
 * without warranty of any kind.                                             *
 *****************************************************************************/

/* $Id: pvfuncs.c,v 1.3 93/08/13 12:25:14 gbourhis Exp $ */

#ifdef RCSLOG
$Log:	pvfuncs.c,v $
 * Revision 1.3  93/08/13  12:25:14  gbourhis
 * add dump_rgb() function. Send data at end of data_load if connected.
 * 
 * Revision 1.2  93/07/13  16:36:35  gbourhis
 * use active_windows field in close_current instead of current_window field.
 * 
 * Revision 1.1  92/09/18  10:55:26  marca
 * Initial revision
 * 
#endif

#include "pv.h"

/* ABORT: */
int abort_actions(state_t *state)
{
  action_t *action;
  action_t *seraction;
  
  /* Free all of the actions. */
  for (seraction = state->action_head; seraction != NULL; seraction = action) 
    {
      action = seraction->next_serial;
      free_action (state, seraction);
    }
  
  /* Reset the state record's pointers to the now empty action list. */
  state->action_head = state->action_tail = NULL;
  
  return ST_OKAY;
}


/* ANIMATE: */
int animate_reverse(state_t *state, window_t *win, int framecount)
{
  if (WIN_FRAMES(win) > 1) 
    {
      WIN_REVERSE(win) = framecount;
      WIN_FORWARD(win) = 0;
    }
  return ST_OKAY;
}


int animate_forward(state_t *state, window_t *win, int framecount)
{
  if (WIN_FRAMES(win) > 1) 
    {
      WIN_FORWARD(win) = framecount;
      WIN_REVERSE(win) = 0;
    }
  return ST_OKAY;
}


int timeblock_forward (state_t *state, window_t *win, 
                       action_t *action, int framecount)
{
  if (WIN_FRAMES(win) > 1)
    {
      if (framecount == -1.0)
        stprintf (state, "Error: Bad framecount.");
      else
        {
          WIN_FORWARD(win) = 0;
          WIN_REVERSE(win) = 0;
          action = add_action
            (state, do_time_forward_step, action->argc, action->arg,
             /* steps */ framecount, /* global */ action->global,
             /* at_current */ TRUE);
          do_time_forward_step (state, action, win);
          (action->countdown)--;
        }
    }
  
  return ST_OKAY;
}

int timeblock_reverse (state_t *state, window_t *win, 
                       action_t *action, int framecount)
{
  if (WIN_FRAMES(win) > 1)
    {
      if (framecount == -1.0)
        stprintf (state, "Error: Bad framecount.");
      else
        {
          WIN_FORWARD(win) = 0;
          WIN_REVERSE(win) = 0;
          action = add_action
            (state, do_time_reverse_step, action->argc, action->arg,
             /* steps */ framecount, /* global */ action->global,
             /* at_current */ TRUE);
          do_time_reverse_step (state, action, win);
          (action->countdown)--;
        }
    }
  
  return ST_OKAY;
}


int do_nothing (state_t *state, action_t *action, window_t *win)
{
  return ST_OKAY;
}


int pause (state_t *state, window_t *win, action_t *action,
           int count)
{
  if (count > 1)
    {
      action = add_action (state, do_nothing, action->argc,
                           action->arg, /* steps */ count,
                           /* global */ action->global, 
                           /* at_current */ TRUE);
      do_nothing (state, action, win);
      (action->countdown)--;
    }
  
  return ST_OKAY;
}


int animate_goto(state_t *state, window_t *win, char *framename)
{
  frame_t *frame;
  
  frame = find_frame_by_name(state, win, framename);
  if (frame == NULL) 
    {
      bprintf(state,
              "WARNING:  Frame %s not in window %s.\n",
              framename, WIN_NAME(win));
      return ST_ERROR;
    }
  else 
    {
      WIN_FRAME(win) = frame;
      WIN_FRAMEID(win) = find_frame_num(state, win, frame);
      animate_stop(state, win);
      set_redraw(state, win);
      send_msg_to_type(state, win, DIALOG, MSG_NEWFRAME);
    }

  return ST_OKAY;
}


int animate_goto_num(state_t *state, window_t *win, int number)
{
  frame_t *frame;
  
  frame = find_frame_by_num(state, win, number);
  if (frame == NULL) 
    {
      bprintf(state,
              "ERROR:  Frame not in window %s.\n", WIN_NAME(win));
      return ST_ERROR;
    }
  else 
    {
      WIN_FRAME(win) = frame;
      WIN_FRAMEID(win) = number;
      animate_stop(state, win);
      set_redraw(state, win);
      send_msg_to_type(state, win, DIALOG, MSG_NEWFRAME);
    }
  
  return ST_OKAY;
}


int animate_stop(state_t *state, window_t *win)
{
  WIN_FORWARD(win) = 0;
  WIN_REVERSE(win) = 0;
  return ST_OKAY;
}


int time_delay(state_t *state, window_t *win, int delay)
{
  WIN_DELAY(win) = delay;
  broadcast_msg(state, win, MSG_RESET);
  return ST_OKAY;
}


int time_skip(state_t *state, window_t *win, int skip)
{
  WIN_SKIP(win) = skip;
  broadcast_msg(state, win, MSG_RESET);
  return ST_OKAY;
}


int animate_reset(state_t *state, window_t *win)
{
  WIN_DELAY(win) = 0;
  WIN_FORWARD(win) = 0;
  WIN_REVERSE(win) = 0;
  WIN_FRAME(win) = WIN_FRAMELIST(win);
  WIN_FRAMEID(win) = 1;
  WIN_ANIMATED(win) = FALSE;
  set_redraw(state, win);
  broadcast_msg(state, win, MSG_RESET);
  
  return ST_OKAY;
}

/* --------------------- COLORMAP LOAD/SAVE ROUTINES ---------------------- */
/* These should get called from file selection box callbacks. */
int cm_load_hdf(state_t *state, window_t *win, char *colormappath)
{
  return ReadHDFPalette(state, win, colormappath);
}


int cm_load_raw(state_t *state, window_t *win, char *colormappath)
{
  return ReadRawPalette(state, win, colormappath);
}


int cm_put_hdf(state_t *state, window_t *win, char *palettepath)
{
  return WriteHDFPalette(state, win, palettepath);
}


int cm_put_raw(state_t *state, window_t *win, char *palettepath)
{
  return WriteRawPalette(state, win, palettepath);
}
/* ------------------------------------------------------------------------ */


int cm_reset(state_t *state, window_t *win)
{
  if (win == NULL)
    stprintf(state, "ERROR:  Can not load palette.  No active window.\n");
  else
    return set_palette(state, win->palette, PAL_RESET);

  return ST_ERROR;
}


int cm_scale(state_t *state, window_t *win, float scalestep, int single)
{
  window_t *palwin;
  palette_t *palette;

  palwin = state->palette_window;
  palette = (palette_t *) WIN_IMAGE(palwin);
  
  if (single) 
    {
      WIN_FORWARD(palwin) = 1;
      WIN_REVERSE(palwin) = 0;
      palette->slope = scalestep;
      palette->slopestep = 0.0;
    }
  else 
    {
      WIN_FORWARD(palwin) = -1;
      WIN_REVERSE(palwin) = 0;
      palette->slopestep = scalestep;
    }
  
  return ST_OKAY;
}


int cm_shift(state_t *state, window_t *win, float shiftstep, int single)
{
  window_t *palwin;
  palette_t *palette;
  
  palwin = state->palette_window;
  palette = (palette_t *) WIN_IMAGE(palwin);
  
  if (single) 
    {
      WIN_FORWARD(palwin) = 1;
      WIN_REVERSE(palwin) = 0;
      palette->origin = shiftstep;
      palette->originstep = 0.0;
    }
  else 
    {
      WIN_FORWARD(palwin) = -1;
      WIN_REVERSE(palwin) = 0;
      palette->originstep = shiftstep;
    }
  
  return ST_OKAY;
}


int cm_load(state_t *state, window_t *win, int colormap)
{
  if (state->active_windows != NULL)
    set_palette(state, WIN_PALETTE(state->active_windows),
                colormap);
  
  return ST_OKAY;
}

/* Dump a POLYVIEW window into RGB format */
int dump_rgb(state_t *state, window_t *win, char *fname)
{
  polyview_t *pv;

  assert (win->type == POLYVIEW);
  pv = (polyview_t *)WIN_IMAGE(win);
  DumpRGB(pv->drawing_volume, WIN_WIDTH(win) - 2, WIN_HEIGHT(win) - 2,
	  fname);
  return ST_OKAY;
}
  

/* HELP: */
int help(state_t *state, window_t *win)
{
  GUIdoHelp (PV_HELP_ONCOMMANDS);
  
  return ST_OKAY;
}



/* SCRIPTING: */
int script(state_t *state, window_t *win, char *filename)
{
  FILE *fp;
    
  /* Open the file. */
  fp = fopen(filename, "r");
  if (fp == NULL) 
    {
      char line[MAXLINELEN];

      sprintf (line, "ERROR: Cannot find script file named %s.", filename);
      XmxMakeErrorDialog (gstate->base, line,
                          "Error While Loading Script");
      XtManageChild (Xmx_w);
      
      bprintf(state,
              "ERROR:  Could not open '%s'.\n", filename);
      return ST_ERROR;
    }
  else 
    {
      stprintf(state, "Loading '%s'...\n", filename);
      load_script(state, fp);
      fclose(fp);
    }
  
  return ST_OKAY;
}

/* ---------------------- MOVEMENT SUPPORT FUNCTIONS ---------------------- */
/* set_camera_pos sets the position of the camera in the window passed */
/* to it.  It updates the spherical coordinates of the window as well.  The */
/* function expects the system state pointer, a pointer to the window, and */
/* the new camera x, y, and z coordinates.  Returns ST_OKAY. */
int set_camera_pos(state_t *state, window_t *win, float x, float y, 
                   float z)
{
  /* Set the win's cartesian coordinates. */
  WIN_CFROM(win)[X] = x;
  WIN_CFROM(win)[Y] = y;
  WIN_CFROM(win)[Z] = z;
  
  /* Set the window's spherical coordinates. */
  cart_to_sphere(x-WIN_AT(win)[X], y-WIN_AT(win)[Y],
                 z-WIN_AT(win)[Z],
                 &WIN_SFROM(win)[RHO], &WIN_SFROM(win)[THETA],
                 &WIN_SFROM(win)[PHI]);
  
  return ST_OKAY;
}


/* Set ref position sets the reference position in the window passed */
/* to it.  It updates the spherical coordinates of the window as well.  The */
/* function expects the system state pointer, a pointer to the window, and */
/* the new reference x, y, and z coordinates.  Returns ST_OKAY. */
int set_ref_pos(state_t *state, window_t *win, float x, float y, float z)
{
  /* Set the window's cartesian coordinates. */
  WIN_AT(win)[X] = x;
  WIN_AT(win)[Y] = y;
  WIN_AT(win)[Z] = z;
  
  /* Set the window's spherical coordinates. */
  cart_to_sphere(WIN_CFROM(win)[X]-x, WIN_CFROM(win)[Y]-y,
                 WIN_CFROM(win)[Z]-z,
                 &WIN_SFROM(win)[RHO], &WIN_SFROM(win)[THETA],
                 &WIN_SFROM(win)[PHI]);
  
  return ST_OKAY;
}


/* This is same as interp_pos below but for twist; accepting starting
   and ending twist values and steps/step as below, return twistp. */
int interp_twist (Angle twist1, Angle twist2,
                  int steps, int step,
                  Angle *twistp)
{
  if ((steps == 0) || (step == steps))
    *twistp = twist2;
  else
    *twistp = twist1 + 
      (Angle)(((float)(twist2-twist1)*(float)step)/(float)steps);

  return ST_OKAY;
}


/* Interp pos interpolates a set of cartesian coordinates to a point along */
/* a parameterized line between the two.  The function takes a starting */
/* x, y, and z location, an ending x, y, and z location, a number of steps */
/* to subdivide the path into, the step for which to return a coordinate, */
/* the type of path to follow (CARTESIAN or SPHERICAL), and the return */
/* value pointers.  Expects coordinates relative to the reference point. */
/* Returns ST_OKAY if no errors occur. */
int interp_pos(float x1, float y1, float z1, float x2, float y2, float z2, 
               int steps, int step, int pathtype, float *xp, float *yp, 
               float *zp)
{
  if ((steps == 0) || (step == steps)) 
    {
      *xp = x2;
      *yp = y2;
      *zp = z2;
    }
  else 
    {
      if (pathtype == CARTESIAN) 
        {
          *xp = x1 + ((x2-x1)*(float)step)/(float)steps;
          *yp = y1 + ((y2-y1)*(float)step)/(float)steps;
          *zp = z1 + ((z2-z1)*(float)step)/(float)steps;
        }
      else if (pathtype == SPHERICAL) 
        {
          float rho1, theta1, phi1;
          float rho2, theta2, phi2;
          float rhop, thetap, phip;
      
          cart_to_sphere(x1, y1, z1, &rho1, &theta1, &phi1);
          cart_to_sphere(x2, y2, z2, &rho2, &theta2, &phi2);
      
          rhop = rho1 + ((rho2-rho1)*(float)step)/(float)steps;
          thetap = theta1 + ((theta2-theta1)*(float)step)/(float)steps;
          phip = phi1 + ((phi2-phi1)*(float)step)/(float)steps;
          
          sphere_to_cart(rhop, thetap, phip, xp, yp, zp);
        }
      else 
        {
          return ST_ERROR;
        }
    }
  return ST_OKAY;
}


/* MOVE: */
int move_camera(state_t *state, window_t *win, action_t *action, int pathtype,
                float to_0, float to_1, float to_2, Angle twist)
{
  int steps;
  float rhodist, thetadist, phidist;
  float x, y, z;
  float rhoavg;
  Angle dtwist;
    
  if (FEQ(to_0, WIN_AT(win)[X]) &&
      FEQ(to_1, WIN_AT(win)[Y]) &&
      FEQ(to_2, WIN_AT(win)[Z])) 
    {
      bprintf(state,
              "WARNING:  Cannot move camera to ref point.\n");
      return ST_ERROR;
    }
  
  if (WIN_STEP(win) <= 0.0) 
    {
      bprintf(state,
              "WARNING:  move step is %.2f\n", WIN_STEP(win));
      return ST_ERROR;
    }
  
  if (pathtype == CARTESIAN) 
    {
      steps = fhypot(WIN_CFROM(win)[X]-to_0,
                     fhypot(WIN_CFROM(win)[Y]-to_1,
                            WIN_CFROM(win)[Z]-to_2)) /
                              (float)(WIN_STEP(win));
    }
  else if (pathtype == SPHERICAL) 
    {
      /* Since the integral of the path is too difficult to */
      /* solve for at the moment, invent a heuristic length */
      /* for the path and use it to determine the number of */
      /* steps. */
      x = to_0 - WIN_AT(win)[X];
      y = to_1 - WIN_AT(win)[Y];
      z = to_2 - WIN_AT(win)[Z];
      cart_to_sphere(x, y, z, &to_0, &to_1, &to_2);
      
      rhodist = ABS(WIN_SFROM(win)[RHO]-to_0);
      rhoavg = (WIN_SFROM(win)[RHO]+to_0) * 0.5;
      thetadist = ABS(WIN_SFROM(win)[THETA]-to_2);
      phidist = ABS(WIN_SFROM(win)[PHI]-to_1);
      dtwist = ABS(WIN_TWIST(win)-twist)/72;

      /* This is one hell of a heuristic. */
      steps = 
        fhypot (rhodist, 
                fhypot(thetadist, phidist)*(rhoavg * 0.5))
            / (float)(WIN_STEP(win));

      /* If we don't have quite enough steps, add a few more. */
      /* We consider 25 steps for 180 degrees always 'enough'. */
      if (steps < dtwist)
        steps = (int) ((float)steps * 1.2);
    }
  else 
    {
      return ST_ERROR;
    }
  
  if (steps <= 0)
    steps = 1;
  
  action = add_action
    (state, do_jump_camera, action->argc, action->arg,
     /*steps*/ steps, /*global*/ action->global, /*at_current*/ TRUE);
  do_jump_camera(state, action, win);
  (action->countdown)--;
  return ST_OKAY;
}



int move_ref_pt(state_t *state, window_t *win, action_t *action, float to_0,
                float to_1, float to_2)
{
  int steps;
    
  if ((to_0 == WIN_CFROM(win)[X]) &&
      (to_1 == WIN_CFROM(win)[Y]) &&
      (to_2 == WIN_CFROM(win)[Z])) 
    {
      bprintf(state,
              "WARNING:  Cannot move ref point to camera.\n");
      return ST_ERROR;
    }
  
  if (WIN_STEP(win) <= 0.0) 
    {
      bprintf(state,
              "WARNING:  move step is %.2f\n", WIN_STEP(win));
      return ST_ERROR;
    }
  
  /* Assume only pathtype is CARTESIAN, for now. */
  steps = fhypot(WIN_AT(win)[X]-to_0,
                 fhypot(WIN_AT(win)[Y]-to_1,
                        WIN_AT(win)[Z]-to_2)) /
                          WIN_STEP(win);
  
  if (steps <= 0)
    steps = 1;
  
  action = add_action
    (state, do_jump_ref_pt, action->argc, action->arg,
     /*steps*/ steps, /*global*/ action->global, /*at_current*/ TRUE);
  do_jump_ref_pt(state, action, win);
  (action->countdown)--;
  return ST_OKAY;
}


float *SelectedCentroid(state_t *state, window_t *win)
{
  static float centroid[DIMS];
  float *coord;
  int nselected;
  object_t *root;
  int i, k;
  object_t *obj;
  pickmode_t pickmode;
    
  /* Initialize the centroid of the selected region and the count of */
  /* selected items. */
  centroid[X] = centroid[Y] = centroid[Z] = 0.0;
  nselected = 0;
  
  /* If nothing is loaded, return a centroid at the origin. */
  if (WinHasNoData(win))
    return centroid;
  
  /* Get the root object of the window. */
  root = WIN_CURROOTOBJ(win);
  
  /* Set up variables for the search. */
  nselected = WIN_SELECTED(win);
  
  /* Get the pick mode of the first selected thing.  This indicates */
  /* which type of selection is active. */
  if (nselected == 0) 
    {
      /* If nothing is selected, assume pickmode is PICK_NONE */
      pickmode = PICK_NONE;
    }
  else 
    {
      obj = GetKthElementObj(root, 1);
      pickmode = OBJ_PICKMODE(obj);
    }
  
  /* Compute the centroid differently depending on what is selected. */
  switch (pickmode) 
    {
    case PICK_VERTEX:
      /* Sum the coordinates of the vertices. */
      for (i = 1; i <= nselected; i++) 
        {
          /* Get the object, element, and corresponding coord. */
          obj = GetKthElementObj(root, i);
          k = GetKthPickedElement(obj, i);
          coord = GetVertexCoord(state, obj, k);
          
          /* Add the coordinate to the total for each member */
          /* of the triple. */
          centroid[X] += coord[X];
          centroid[Y] += coord[Y];
          centroid[Z] += coord[Z];
        }
      break;
    case PICK_FACE:
      /* Sum the centroids of the faces. */
      for (i = 1; i <= nselected; i++) 
        {
          /* Get the object, element, and corresponding coord. */
          obj = GetKthElementObj(root, i);
          k = GetKthPickedElement(obj, i);
          coord = GetFaceCoord(state, obj, k);
          
          /* Add the coordinate to the total for each member */
          /* of the triple. */
          centroid[X] += coord[X];
          centroid[Y] += coord[Y];
          centroid[Z] += coord[Z];
        }
      break;
    case PICK_BELOW:
    case PICK_OBJECT:
      /* Sum the centroids of the selected objects. */
      for (i = 1; i <= nselected; i++) 
        {
          /* Get the object. */
          obj = GetKthElementObj(root, i);
          
          /* Add the coordinate to the total for each member */
          /* of the triple. */
          centroid[X] += OBJ_STATS(obj, PCOORD, X).mean;
          centroid[Y] += OBJ_STATS(obj, PCOORD, Y).mean;
          centroid[Z] += OBJ_STATS(obj, PCOORD, Z).mean;
        }
      break;
    case PICK_NONE:
    case PICK_ALL:
      /* Take the mean of the root object's contents. */
      centroid[X] = OBJ_STATS(obj, PCOORD, X).mean;
      centroid[Y] = OBJ_STATS(obj, PCOORD, Y).mean;
      centroid[Z] = OBJ_STATS(obj, PCOORD, Z).mean;
      break;
    default:
      /* Otherwise, return the origin. */
      return centroid;
    }
  
  /* Calculate the centroid of the selected items. */
  centroid[X] /= (float)nselected;
  centroid[Y] /= (float)nselected;
  centroid[Z] /= (float)nselected;
  return centroid;
}


int move_ref_sel(state_t *state, window_t *win, action_t *action)
{
  float *centroid;
  char value[MAXLINELEN];
  
  /* Get the centroid of the selected elements and move the focus */
  /* point to there. */
  centroid = SelectedCentroid(state, win);
  sprintf(value, "move focus point %f %f %f", centroid[X], centroid[Y],
          centroid[Z]);
  return parse_line(state, value, TRUE);
}


int move_path(state_t *state, window_t *win, int movepath)
{
  WIN_MOVEPATH(win) = movepath;
  return ST_OKAY;
}


int move_step(state_t *state, window_t *win, float step)
{
  WIN_STEP(win) = step;
  return ST_OKAY;
}


/* JUMP: */
int jump_camera(state_t *state, window_t *win, int steps,
                float to_0, float to_1, float to_2, Angle twist)
{
  /* Starting point is WIN_CFROM-WIN_AT.
     Ending point is to-WIN_AT.
     Steps is 'steps'.
     1 is 'step' (number of steps in this step, apparently
       this is always 1.
     WIN_MOVEPATH is sph or cart.
     &WIN_CFROM is position after incrementing this time. */

  /* So starting point is (eyepoint - refpoint).
     Ending point is (desiredeyepoint - refpoint).
     Result is new eyepoint, to which refpoint must be added. */
  interp_pos(WIN_CFROM(win)[X]-WIN_AT(win)[X],
             WIN_CFROM(win)[Y]-WIN_AT(win)[Y],
             WIN_CFROM(win)[Z]-WIN_AT(win)[Z],
             to_0-WIN_AT(win)[X],
             to_1-WIN_AT(win)[Y],
             to_2-WIN_AT(win)[Z],
             steps, 1, WIN_MOVEPATH(win),
             &WIN_CFROM(win)[X],
             &WIN_CFROM(win)[Y],
             &WIN_CFROM(win)[Z]);

  /* Go interpolate the twist now too. */
  interp_twist (WIN_TWIST(win), twist, steps, 1, &WIN_TWIST(win));
  win->twist_tmp = WIN_TWIST(win);

  /* Add cartesian location of reference point to eyepoint. */
  WIN_CFROM(win)[X] += WIN_AT(win)[X];
  WIN_CFROM(win)[Y] += WIN_AT(win)[Y];
  WIN_CFROM(win)[Z] += WIN_AT(win)[Z];

  /* All this does is set up WIN_SFROM for the window. */
  set_camera_pos(state, win,
                 WIN_CFROM(win)[X],
                 WIN_CFROM(win)[Y],
                 WIN_CFROM(win)[Z]);

  /* Redraw the window and we're outta here. */
  set_redraw(state, win);
  return ST_OKAY;
}


int jump_ref_pt(state_t *state, window_t *win, int steps, 
                float to_0, float to_1, float to_2)
{
  interp_pos(WIN_AT(win)[X], WIN_AT(win)[Y], WIN_AT(win)[Z],
             to_0, to_1, to_2,
             steps, 1, WIN_MOVEPATH(win),
             &WIN_AT(win)[X], &WIN_AT(win)[Y], &WIN_AT(win)[Z]);
  
  set_ref_pos(state, win,
              WIN_AT(win)[X],
              WIN_AT(win)[Y],
              WIN_AT(win)[Z]);
  
  set_redraw(state, win);
  return ST_OKAY;
}


int jump_ref_sel(state_t *state, window_t *win)
{
  float *centroid;
  
  /* Get the centroid of the selected elements and jump the focus */
  /* point to there. */
  centroid = SelectedCentroid(state, win);
  return jump_ref_pt(state, win, 1,
                     centroid[X], centroid[Y], centroid[Z]);
}


/* Overlays: */
int set_overlay(state_t *state, window_t *win, int type, int style, 
                int hereonly)
{
  int *value;
  
  if (hereonly)
    value = &(WIN_ROOTOBJINFO(win)->hereonly.value);
  else
    value = &(WIN_ROOTOBJINFO(win)->heredown.value);
  
  if (style == TOGGLE) 
    {
      *value = *value ^ type;
      set_redraw(state, win);
    }
  else 
    {
      if ((*value & type) != style) 
        {
          *value = (*value & (~type)) | style;
          set_redraw(state, win);
        }
    }

  GUIpviewResetProps (state, win);

  return ST_OKAY;
}

int set_over_shadow_vertices(state_t *state, window_t *win)
{
  WIN_SCATDRAW(win) = draw_vertices;
  set_redraw(state, win);
  return ST_OKAY;
}


int set_over_shadow_lines(state_t *state, window_t *win)
{
  WIN_SCATDRAW(win) = draw_edges;
  set_redraw(state, win);
  return ST_OKAY;
}


int set_over_shadow_solid(state_t *state, window_t *win)
{
  WIN_SCATDRAW(win) = draw_faces;
  set_redraw(state, win);
  return ST_OKAY;
}


/* Project: */
int project_ortho(state_t *state, window_t *win)
{
  if (WIN_PROJECTION(win) != ORTHOGRAPHIC) 
    {
      WIN_PROJECTION(win) = ORTHOGRAPHIC;
      set_redraw(state, win);
    }
  
  GUIpviewResetRender (state, win);
  
  return ST_OKAY;
}


int project_perspective(state_t *state, window_t *win)
{
  if (WIN_PROJECTION(win) != PERSPECTIVE) 
    {
      WIN_PROJECTION(win) = PERSPECTIVE;
      set_redraw(state, win);
    }
  
  GUIpviewResetRender (state, win);
  
  return ST_OKAY;
}


/* QUIT: */
int quit(state_t *state)
{
  exit_polyview ();
}



/* DRAW: */
static long SetDrawFunction(state_t *state, object_t *object, va_list args)
{
  int (*fn)();
  
  if (OBJ_PICKMODE(object) != PICK_NONE) 
    {
      fn = (int (*)()) va_arg(args, long);
      object->info->draw_fn = fn;
    }
  
  return 0;
}


int set_draw_nothing(state_t *state, window_t *win)
{
  if (win->frame == NULL) 
    {
      WIN_DRAW(win) = draw_nothing;
      GUIpviewResetRender (state, win);
      return ST_OKAY;
    }
    
  ApplyFuncToTree(state, WIN_ROOTOBJ(win), SetDrawFunction,
                  draw_nothing);
  set_redraw(state, win);
  
  GUIpviewResetRender (state, win);
  
  return ST_OKAY;
}

int set_draw_vertices(state_t *state, window_t *win)
{
  if (win->frame == NULL) 
    {
      WIN_DRAW(win) = draw_vertices;
      GUIpviewResetRender (state, win);
      return ST_OKAY;
    }
    
  ApplyFuncToTree(state, WIN_ROOTOBJ(win), SetDrawFunction,
                  draw_vertices);
  set_redraw(state, win);
  
  GUIpviewResetRender (state, win);
  
  return ST_OKAY;
}


int set_draw_lines(state_t *state, window_t *win)
{
  if (win->frame == NULL) 
    {
      WIN_DRAW(win) = draw_edges;
      GUIpviewResetRender (state, win);
      return ST_OKAY;
    }
    
  ApplyFuncToTree(state, WIN_ROOTOBJ(win), SetDrawFunction,
                  draw_edges);
  set_redraw(state, win);
  
  GUIpviewResetRender (state, win);
  
  return ST_OKAY;
}


int set_draw_hidden_edges(state_t *state, window_t *win)
{
  if (win->frame == NULL) 
    {
      WIN_DRAW(win) = draw_hidden_edges;
      GUIpviewResetRender (state, win);
      return ST_OKAY;
    }
    
  ApplyFuncToTree(state, WIN_ROOTOBJ(win), SetDrawFunction,
                  draw_hidden_edges);
  set_redraw(state, win);
  
  GUIpviewResetRender (state, win);
  
  return ST_OKAY;
}


int set_draw_transparent(state_t *state, window_t *win)
{
  if (win->frame == NULL) 
    {
      WIN_DRAW(win) = DrawTrans;
      GUIpviewResetRender (state, win);
      return ST_OKAY;
    }
    
  ApplyFuncToTree(state, WIN_ROOTOBJ(win), SetDrawFunction, DrawTrans);
  set_redraw(state, win);
  
  GUIpviewResetRender (state, win);
  
  return ST_OKAY;
}


int set_draw_solid(state_t *state, window_t *win)
{
  if (win->frame == NULL) 
    {
      WIN_DRAW(win) = draw_faces;
      GUIpviewResetRender (state, win);
      return ST_OKAY;
    }
  
  ApplyFuncToTree(state, WIN_ROOTOBJ(win), SetDrawFunction,
                  draw_faces);
  set_redraw(state, win);
  
  GUIpviewResetRender (state, win);
  
  return ST_OKAY;
}


int set_draw_outlined(state_t *state, window_t *win)
{
  if (win->frame == NULL) 
    {
      WIN_DRAW(win) = draw_outlined_faces;
      GUIpviewResetRender (state, win);
      return ST_OKAY;
    }
    
  ApplyFuncToTree(state, WIN_ROOTOBJ(win), SetDrawFunction,
                  draw_outlined_faces);
  set_redraw(state, win);
  
  GUIpviewResetRender (state, win);
  
  return ST_OKAY;
}


int set_draw_fast(state_t *state, window_t *win, int (*fast_fn)())
{
  if (WIN_FRAMES(win) >= 1)
    WIN_FASTDRAW_FN(win) = fast_fn;
  
  GUIpviewResetRender (state, win);
  
  return ST_OKAY;
}



/* RESET: */
int reset(state_t *state, window_t *win)
{
  int data_scale(state_t *, window_t *, float, float, float);
  
  init_window(state, win, TRUE);
  data_scale(state, win, 1.0, 1.0, 1.0);
  
  GUIpviewResetRender (state, win);

  return ST_OKAY;
}



/* DATA: */
int data_clear(state_t *state, window_t *win)
{
  frame_t *frame;
    
  /* Recursively delete the frames and objects associated with this */
  /* win. */
  frame = get_first_frame(state, win);
  while (frame != NULL)
    frame = delete_frame(state, win, frame);
  
  /* Remover the links from the windows to the deleted frame. */
  WIN_FRAME(win) = WIN_FRAMELIST(win) = NULL;
  WIN_FRAMES(win) = 0;
  
  return ST_OKAY;
}


int data_file(state_t *state, window_t *win, char *filename)
{
  if ((WIN_SOURCE(win) = find_file(state, HDFVSET, filename)) == NULL)
    {
      char line[MAXLINELEN];

      sprintf (line, "ERROR: Cannot find file named %s.", filename);
      XmxMakeErrorDialog (gstate->base, line,
                          "Error While Finding File");
      XtManageChild (Xmx_w);
      return ST_ERROR;
    }
  else
    return ST_OKAY;
}


int data_dtm(state_t *state, window_t *win, char *filename)
{
  if ((WIN_SOURCE(win) = find_file(state, DTMPORT, filename)) == NULL)
    return ST_ERROR;
  else
    return ST_OKAY;
}

int data_dtm_explicit (state_t *state, window_t *win, char *filename)
{
  WIN_SOURCE(win) = set_file_dtm_explicit (state, filename);
  return ST_OKAY;
}


int data_frame(state_t *state, window_t *win, char *framename)
{
  /* Find the group with the given name. */
  if ((WIN_GROUP(win) = find_vgroup(state, WIN_SOURCE(win), framename))
      == NULL)
    /* No such group */
    return ST_ERROR;
  
  strcpy(WIN_GROUPNAME(win), framename);
  return ST_OKAY;
}


int set_data_name(state_t *state, action_t *action, window_t *win, int slot,
                  char *defname)
{
  char *name;
  
  name = get_string_arg(action->argc, action->arg, "name", NULL);
  if (name == NULL)
    strcpy(win->vname[slot], defname);
  else
    strcpy(win->vname[slot], name);
  
  return ST_OKAY;
}


int data_load(state_t *state, window_t *win, int mode)
{
  Vgroup_t *group;
  frame_t *frame;
  
  /* Get a pointer to the group. */
  group = WIN_GROUP(win);
  if (group == NULL)
    {
      GUIsetStatus ("Doing data frame /* automatically.");
      data_frame (state, win, "/*");
      group = WIN_GROUP(win);
    }

  stprintf (state, "Loading data...");

  /* Free any existing frames from the window. */
  frame = get_first_frame(state, win);
  while (frame != NULL)
    frame = delete_frame(state, win, frame);
  WIN_FRAMELIST(win) = NULL;
  
  /* Now that we have a group, recursively build the frames */
  /* and objects that it represents. */
  if (build_frames(state, win, group) == ST_ERROR)
    {
      stprintf (state, "Could not load data.");
      return ST_ERROR;
    }

  /* Initialize the win with the data. */
  init_window(state, win, mode);

  if (WIN_SOURCE(win)->type ==  HDFVSET && COMdtm_out_active_p (state))
    {
      COMsend_vset (state, group);
      NetSendCommand
	(NULL, "Polyview", "polyview data load", NULL, NULL);
    }

  stprintf (state, "Data loaded.");
  
  broadcast_msg(state, win, MSG_NEWDATA);
  
  return ST_OKAY;
}


int data_scale (state_t *state, window_t *win, float xscale, float yscale, 
                float zscale)
{
  objinfo_t *info;
  float xratio, yratio, zratio;
  
  /* Only apply the scaling if the window contains data. */
  if (WinHasData(win) && (WIN_ROOTOBJINFO(win) != NULL)) 
    {
      /* Load root object info record with scale values. */
      info = WIN_ROOTOBJINFO(win);
      info->scl.used = TRUE;
      xratio = xscale / info->scl.coord[X];
      yratio = yscale / info->scl.coord[Y];
      zratio = zscale / info->scl.coord[Z];
      info->scl.coord[X] = xscale;
      info->scl.coord[Y] = yscale;
      info->scl.coord[Z] = zscale;
      
      /* Adjust the maxwidth. */
      WIN_MAXWIDTH(win) =
        fhypot((WIN_STATS(win,PX).max-WIN_STATS(win,PX).min)*xscale,
               fhypot((WIN_STATS(win,PY).max-WIN_STATS(win,PY).min)*yscale,
                      (WIN_STATS(win,PZ).max-WIN_STATS(win,PZ).min)* zscale));
      
      /* Set the window to redraw. */
      set_redraw(state, win);
    }
  else 
    {
      /* Tell the user why nothing was done. */
      bprintf(state, "WARNING: Cannot scale, no data.\n");
    }
  
  return ST_OKAY;
}


int data_normalize(state_t *state, window_t *win)
{
  return data_scale(state, win,
                    1.0/(WIN_STATS(win,PX).max-WIN_STATS(win,PX).min),
                    1.0/(WIN_STATS(win,PY).max-WIN_STATS(win,PY).min),
                    1.0/(WIN_STATS(win,PZ).max-WIN_STATS(win,PZ).min));
}


/* SELECT: */
int sel_all(state_t *state, window_t *win)
{
  /* Select all of the objects in the window. */
  PickAll(state, win);
  
  /* Update the window to show the new picking state. */
  set_redraw(state, win);
  return ST_OKAY;
}


int sel(state_t *state, window_t *win, pickmode_t mode)
{
  win->pickmode = mode;

  GUIpviewResetSelect (state, win);

  return sel_all(state, win);
}


/* FUCK FUCK FUCK! */
int sel_region(state_t *state, window_t *win, short left, short right,
               short bottom, short top, int toggle)
{
  PickRegion(state, win, left, bottom, right, top, toggle, 1);
  return ST_OKAY;
}



/* WINDOWS: */
int win_geometry(state_t *state, long left, long bottom, long right, long top)
{
  if ((left >= right) && (bottom >= top)) {
    return ST_ERROR;
  }
  else {
    /* prefposition(left, right, bottom, top); */
    win_geometry_auto(state, PREF);
    return ST_OKAY;
  }
}

int win_geometry_auto(state_t *state, int mode)
{
  if (mode == PREF) 
    {
      if (state->geometry < PREF) 
        state->geometry += PREF;
    }
  else 
    {
      state->geometry = mode;
    }
  return ST_OKAY;
}

int win_open_palette(state_t *state)
{
  post_palette(state);
  return ST_OKAY;
}


int win_activate(state_t *state, window_t *win)
{
  window_t *w;

  broadcast_msg(state, win, MSG_NEWACT, state->active_windows);
  state->active_windows = win;

  /* Adjust 'active' buttons on all windows! */
  FOR_ALL_WINDOWS(state, w)
    if (w->type == POLYVIEW)
      XmxSetToggleButton 
        (((polyview_t *)WIN_IMAGE(w))->active, (win == w) ? XmxSet : XmxUnset);

  /* Right now this won't result in anything written to the
     transcript, ever. */
  /* Maybe we should always make sure a 'window activate' command
     is written to the transcript whenever a _new_ window is made
     active... */
  
  return ST_OKAY;
}


int win_close_current(state_t *state)
{
  window_t *win;

  /* Is this OK? */
  win = state->active_windows;
  if (win)
    delete_window(state, win);
  state->active_windows = NULL;
  return ST_OKAY;
}


int win_close_palette(state_t *state)
{
  if (STA_PALETTE(state) != NULL)
    XtUnmanageChild (STA_PALETTE(state)->base);
  return ST_OKAY;
}


int win_close_time(state_t *state)
{
  if (state->animate_window != NULL)
    XtUnmanageChild (state->animate_window->base);
  return ST_OKAY;
}


int win_close_info(state_t *state, window_t *win, char *name)
{
  destroy_info (state, win);
  return ST_OKAY;
}


int win_close_polyview(state_t *state, window_t *win, char *name)
{
  win = find_window_by_name(state, name, POLYVIEW);
  if (win)
    delete_window(state, win);
  else
    stprintf(state, "can't find window named \"%s\"", name);
  return ST_OKAY;
}
