/*****************************************************************************
 * NCSA Polyview 3.1                                                         *
 *                                                                           *
 * Version 3.1 changes and additions by Gilles Bourhis.                      *
 * Version 3 changes and additions by Marc Andreessen.                       *
 * Version 2 by Brian Calvert.                                               *
 *                                                                           *
 * Software Development Group                                                *
 * National Center for Supercomputing Applications                           *
 * University of Illinois at Urbana-Champaign                                *
 *                                                                           *
 * This is BETA release software.  As such it may contain software bugs and  *
 * exhibit inconsistencies.                                                  *
 *                                                                           *
 * Please send bug reports to polyview@ncsa.uiuc.edu.                        *
 *                                                                           *
 * Copyright (c) 1992 The Board of Trustees of the University of Illinois.   *
 *                                                                           *
 * Permission to use, copy, and modify this software and its                 *
 * documentation for educational, research, and non-profit purposes is       *
 * hereby granted, provided that the above copyright notice, the original    *
 * authors names, and this permission notice appear in all such copies.      *
 * Any distribution of this software requires the explicit and written       *
 * authorization of the authors.                                             *
 *                                                                           *
 * The University of Illinois makes no representations about the             *
 * suitability of this software for any purpose.  It is provided "as is"     *
 * without warranty of any kind.                                             *
 *****************************************************************************/

/* $Id: pvstatus.c,v 1.3 93/08/24 10:27:03 gbourhis Exp $ */

#ifdef RCSLOG
$Log:	pvstatus.c,v $
 * Revision 1.3  93/08/24  10:27:03  gbourhis
 * replace 3.0 by 3.1.
 * 
 * Revision 1.2  92/09/20  11:32:14  marca
 * Removed BETA in status msg.
 * 
 * Revision 1.1  1992/09/18  10:55:26  marca
 * Initial revision
 *
#endif

#include "pv.h"

/* init status initializes the status window's data structures. 
   Returns ST_OKAY. */
static int init_status(state_t *state, window_t *win, int reset)
{
  /* Nothing happens here. */
  return ST_OKAY;
}

static int redraw_status(state_t *state, window_t *win)
{
  /* This should never be called. */
  return ST_OKAY;
}

static int animate_status(state_t *state, window_t *win)
{
  /* This should never be called either. */
  return ST_OKAY;
}

static int event_status(state_t *state, window_t *win, long dev_id, short data)
{
  /* This is wholly irrelevant. */
  return ST_OKAY;
}

static int notify_status (state_t *state, window_t *send_win, 
                          window_t *recv_win, int message, va_list args)
{
  int status;
  
  status = ST_OKAY;
  
  /* Status windows don't care about messages. */
  switch (message) 
    {
    case MSG_NEWACT:
    case MSG_NEWFRAME:
    case MSG_NEWDATA:
    case MSG_NEWPICK:
    case MSG_VIEWCHG:
    case MSG_CLOSING:
    case MSG_RESET:
    default:
      break;
    }
  
  return status;
}

static int destroy_status(state_t *state, window_t *win)
{
  /* This should shut down the dialog; however, this will
     never get called for the main window, right? */
  return ST_OKAY;
}

XmxCallback (GUItimeoutOne)
{
  /* Add timeout again. */
  XtAppAddTimeOut (gstate->app_context, 50, GUItimeoutOne, True);

  return;
}

XmxCallback (GUItimeoutTwo)
{
  /* This used to be up above. */
  animate_windows (gstate);

  process_next_command (gstate);

  COMpoll (gstate);

  process_events (gstate);

  /* Add timeout again. */
  XtAppAddTimeOut (gstate->app_context, 50, GUItimeoutTwo, True);

  return;
}


/* ---------------------------- GUIaddCommand ----------------------------- */
/* Add a command string to the command history list. */
void GUIaddCommand (char *command_str)
{
  window_t *win = gstate->status;
  status_t *st = (status_t *)(win->image);
  Widget list = st->command_list;

  XmListAddItemUnselected (list, XmxMakeXmstrFromString (command_str), 0);
  XmListSetBottomPos (list, 0);

  /* We assume that if we're passing through here then the command
     wasn't processed via parse_line, so we should send it over the
     net explicitly. */
  COMsend_com (gstate, command_str);

  return;
}

/* ----------------------------- GUIsetStatus ----------------------------- */
void GUIsetStatus (char *str)
{
  window_t *win;
  status_t *st; 

  win = gstate->status;
  assert (win != NULL);
  st = (status_t *)WIN_IMAGE(win);

  XmxTextSetString (st->status, str);
  return;
}


/* --------------------- CALLBACKS FOR STATUS WINDOW ---------------------- */
XmxCallback (GUIcommandCB)
{
  char *c;
  PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/click3.aiff");

  /* c = (char *)PVMALLOC (128 * sizeof (char)); */

  XmStringGetLtoR 
    ((((XmCommandCallbackStruct *)call_data)->value),
     XmSTRING_DEFAULT_CHARSET, &c);

  /* Warn parse_line that this command is already in
     the GUI history list. */
  gstate->last_cmd_already_added = TRUE;

  /* Do something with this command. */
  parse_line (gstate, c, TRUE);
  
  COMsend_com (gstate, c);

  return;
}

XmxCallback (GUIexitConfirmationCB)
{
  /* We get 1 for yes, 0 for no. */
  if (XmxExtractToken ((int)client_data))
    {
      PLAYAFILE ("/usr/lib/sounds/prosonus/musictags/tag2.aiff");
      exit_polyview ();
    }
  else
    {
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/tennis.aiff");
      XtUnmanageChild (w);
    }
  
  return;
}


/* ------------------------- GUIscriptFSBCallback ------------------------- */
XmxCallback (GUIscriptFSBCallback)
{
  char *fname, *msg;
  FILE *fp;

  PLAYAFILE ("/usr/lib/sounds/prosonus/instr/hat_open_closed.aiff");
  
  fname = (char *)PVMALLOC (128 * sizeof (char));
  msg = (char *)PVMALLOC (200 * sizeof (char));
  XmStringGetLtoR 
    (((XmFileSelectionBoxCallbackStruct *)call_data)->value,
     XmSTRING_DEFAULT_CHARSET, &fname);

  fp = fopen (fname, "r");
  if (fp == NULL)
    {
      char line[MAXLINELEN];

      sprintf (line, "ERROR: Cannot find script file named %s.", fname);
      XmxMakeErrorDialog (gstate->base, line,
                          "Error While Loading Script");
      XtManageChild (Xmx_w);
      
      sprintf (msg, "Could not load script %s.\0", fname);
      GUIsetStatus (msg);
    }
  else      
    {
      sprintf (msg, "Loading script %s.\0", fname);
      GUIsetStatus (msg);
      load_script (gstate, fp);
      fclose (fp);
    }
  
  PVFREE (fname);
  PVFREE (msg);

  /* Don't forget to unmanage the FSB. */
  XtUnmanageChild (w);

  return;
}


/* ------------------------------------------------------------------------ */
/* ----------------------------- MAIN MENUBAR ----------------------------- */
/* ------------------------------------------------------------------------ */

/* Hahahahahaha merry little hacks. */
static int do_dither = 1;
static Widget dtmin_box = NULL, dtmin_text;
static Widget dtmout_box = NULL, dtmout_text;


/* ------------------------------------------------------------------------ */
/* -------------------------------- DTM IN -------------------------------- */
/* ------------------------------------------------------------------------ */

static XmxCallback (GUIdtmInCB)
{
  char line[MAXLINELEN];

  /* If client_data is 1, activate the port, else just return. */
  if (!XmxExtractToken ((int)client_data))
    goto done;

  sprintf (line, "dtm in %s\0", XmxTextGetString (dtmin_text));
  parse_line (gstate, line, TRUE);

 done:
  XtUnmanageChild (dtmin_box);
  return;
}


static void GUImakeDTMInBox (state_t *state)
{
  Widget dialog_frame, control_form, in_label;
  Widget dialog_sep, buttons_form;

  /* Zero el uniqid. */
  XmxSetUniqid (0);

  dtmin_box = XmxMakeFormDialog (state->base, "DTM Input Port");
  dialog_frame = XmxMakeFrame (dtmin_box, XmxShadowOut);
  
  XmxSetConstraints 
    (dialog_frame, XmATTACH_FORM, XmATTACH_FORM, XmATTACH_FORM, XmATTACH_FORM,
     NULL, NULL, NULL, NULL);

  /* Main control form. */
  XmxSetArg (XmNhorizontalSpacing, 10);
  control_form = XmxMakeForm (dialog_frame);

  /* Label and text. */
  in_label = XmxMakeLabel (control_form, "DTM Input Port:");
  XmxSetArg (XmNcolumns, 20);
  dtmin_text = XmxMakeText (control_form);
  XmxAddCallbackToText (dtmin_text, GUIdtmInCB, 1);

  /* Separator and buttons for dialog box. */
  dialog_sep = XmxMakeHorizontalSeparator (control_form);
  buttons_form = XmxMakeFormAndTwoButtons
    (control_form, GUIdtmInCB, "Activate", "Cancel", 1, 0);
  
  /* Constraints for control_form. */
  XmxSetArg (XmNtopOffset, 10);
  XmxSetConstraints 
    (in_label, XmATTACH_FORM, XmATTACH_NONE, XmATTACH_FORM, XmATTACH_NONE,
     NULL, NULL, NULL, NULL);
  XmxSetArg (XmNbottomOffset, 10);
  XmxSetConstraints
    (dtmin_text, XmATTACH_WIDGET, XmATTACH_NONE, XmATTACH_FORM,
     XmATTACH_FORM, in_label, NULL, NULL, NULL);
  XmxSetArg (XmNtopOffset, 10);
  XmxSetConstraints
    (dialog_sep, XmATTACH_WIDGET, XmATTACH_WIDGET, XmATTACH_FORM,
     XmATTACH_FORM, dtmin_text, buttons_form, NULL, NULL);
  XmxSetConstraints
    (buttons_form, XmATTACH_NONE, XmATTACH_FORM, XmATTACH_FORM, XmATTACH_FORM,
     NULL, NULL, NULL, NULL);

  return;
}

/* ------------------------------------------------------------------------ */
/* ------------------------------- DTM OUT -------------------------------- */
/* ------------------------------------------------------------------------ */

static XmxCallback (GUIdtmOutCB)
{
  char line[MAXLINELEN];

  /* If client_data is 1, activate the port, else just return. */
  if (!XmxExtractToken ((int)client_data))
    goto done;

  sprintf (line, "dtm out %s\0", XmxTextGetString (dtmout_text));
  parse_line (gstate, line, TRUE);

 done:
  XtUnmanageChild (dtmout_box);
  return;
}


static void GUImakeDTMOutBox (state_t *state)
{
  Widget dialog_frame, control_form, out_label;
  Widget dialog_sep, buttons_form;

  /* Zero el uniqid. */
  XmxSetUniqid (0);

  dtmout_box = XmxMakeFormDialog (state->base, "DTM Output Port");
  dialog_frame = XmxMakeFrame (dtmout_box, XmxShadowOut);
  
  XmxSetConstraints 
    (dialog_frame, XmATTACH_FORM, XmATTACH_FORM, XmATTACH_FORM, XmATTACH_FORM,
     NULL, NULL, NULL, NULL);

  /* Main control form. */
  XmxSetArg (XmNhorizontalSpacing, 10);
  control_form = XmxMakeForm (dialog_frame);

  /* Label and text. */
  out_label = XmxMakeLabel (control_form, "DTM Output Port:");
  XmxSetArg (XmNcolumns, 20);
  dtmout_text = XmxMakeText (control_form);
  XmxAddCallbackToText (dtmout_text, GUIdtmOutCB, 1);

  /* Separator and buttons for dialog box. */
  dialog_sep = XmxMakeHorizontalSeparator (control_form);
  buttons_form = XmxMakeFormAndTwoButtons
    (control_form, GUIdtmOutCB, "Activate", "Cancel", 1, 0);
  
  /* Constraints for control_form. */
  XmxSetArg (XmNtopOffset, 10);
  XmxSetConstraints 
    (out_label, XmATTACH_FORM, XmATTACH_NONE, XmATTACH_FORM, XmATTACH_NONE,
     NULL, NULL, NULL, NULL);
  XmxSetArg (XmNbottomOffset, 10);
  XmxSetConstraints
    (dtmout_text, XmATTACH_WIDGET, XmATTACH_NONE, XmATTACH_FORM,
     XmATTACH_FORM, out_label, NULL, NULL, NULL);
  XmxSetArg (XmNtopOffset, 10);
  XmxSetConstraints
    (dialog_sep, XmATTACH_WIDGET, XmATTACH_WIDGET, XmATTACH_FORM,
     XmATTACH_FORM, dtmout_text, buttons_form, NULL, NULL);
  XmxSetConstraints
    (buttons_form, XmATTACH_NONE, XmATTACH_FORM, XmATTACH_FORM, XmATTACH_FORM,
     NULL, NULL, NULL, NULL);

  return;
}

/* ------------------------ MAIN MENUBAR CALLBACK ------------------------- */

/* This should be public and externally declared so other menubars
   can use it. */
XmxCallback (GUImainMenubarCB)
{
  window_t *win;

  switch (XmxExtractToken ((int)client_data))
    {
    case PV_LOAD_SCRIPT:
      {
        status_t *st;
        
        /* Go get the status record for the main window. */
        win = gstate->status;
        st = (status_t *)WIN_IMAGE(win);
        
        if (st->load_script == NULL)
          st->load_script = XmxMakeFileSBDialog
            (win->base, "Load Polyview Script", 
             "Enter filename of Polyview script:",
             GUIscriptFSBCallback, PV_LOAD_SCRIPT);
        else
          XmFileSelectionDoSearch (st->load_script, NULL);
        XmxManageRemanage (st->load_script);
      }
      break;
    case PV_EXIT_PROGRAM:
      if (gstate->exitbox == NULL)
        {
          gstate->exitbox = XmxMakeQuestionDialog 
            (gstate->base, "Are you sure you want to exit Polyview?",
             "Exit Confirmation", GUIexitConfirmationCB, 1, 0);
          XtManageChild (gstate->exitbox);
        }
      else
        {
          XmxManageRemanage (gstate->exitbox);
        }
      break;

    case PV_DTM_IN:
      if (dtmin_box == NULL)
        {
          GUImakeDTMInBox (gstate);
          XtManageChild (dtmin_box);
        }
      else
        {
          XmxManageRemanage (dtmin_box);
        }
      break;
    case PV_DTM_OUT:
      if (dtmout_box == NULL)
        {
          GUImakeDTMOutBox (gstate);
          XtManageChild (dtmout_box);
        }
      else
        {
          XmxManageRemanage (dtmout_box);
        }
      break;

    case PV_OPEN_POLYVIEW:
      parse_line (gstate, "window open polyview", TRUE);
      break;
    case PV_OPEN_PALETTE:
      parse_line (gstate, "window open palette", TRUE);
      break;
    case PV_OPEN_INFO:
      parse_line (gstate, "window open info", TRUE);
      break;
    case PV_OPEN_TIME:
      parse_line (gstate, "window open time", TRUE);
      break;
    case PV_AUTOPLACE:
      break;

#ifdef DO_AUDIO
    case PV_TOGGLE_SOUND:
      if (gstate->can_use_sound)
        {
          gstate->use_sound = 1 - gstate->use_sound;
          if (gstate->use_sound)
            PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/click3.aiff");
        }
      break;
#endif

    /* Only SGI GL supports dithering right now. */
#ifdef __sgi
    case PV_TOGGLE_DITHER:
      /* Using dither() when GL isn't active coredumps. */
      if (gstate->gl_initialized)
        {
          do_dither = 1 - do_dither;
          FOR_ALL_WINDOWS (gstate, win)
            if (win->type == POLYVIEW)
              {
                polyview_t *pv = (polyview_t *)WIN_IMAGE(win);
                XmxWinset (pv->drawing_volume);

                if (do_dither)
                  dither (DT_ON);
                else
                  dither (DT_OFF);
                set_redraw (gstate, win);
              }

          PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/click2.aiff");
        }
      else
        {
          /* Force it to stay at original value. */
          XmxRSetToggleState 
            (((status_t *)(gstate->status->image))->menubar, PV_TOGGLE_DITHER,
             (do_dither) ? XmxSet : XmxUnset);
        }
      break;
#endif /* __sgi */

    case PV_HELP_ABOUT:
      GUIdoHelp (PV_HELP_ABOUT);
#if 0
      XmUpdateDisplay (gstate->base);
      PLAYAFILE ("pvintro.aiff");
#endif
      break;
    case PV_HELP_ONCOMMANDS:
      GUIdoHelp (PV_HELP_ONCOMMANDS);
      break;
    case PV_MAIN_HELP_ONWINDOW:
      GUIdoHelp (PV_MAIN_HELP_ONWINDOW);
#if 0
      XmUpdateDisplay (gstate->base);
      PLAYAFILE ("pvstatus.aiff");
#endif
      break;
    }
  
  return;
}


/* ---------------------------- MENUBAR STRUCT ---------------------------- */
static XmxMenubarStruct FileMenu[] =
{
  { "Load Script...",  'L', GUImainMenubarCB, PV_LOAD_SCRIPT },
  { "----" },
  { "DTM In...",       'I', GUImainMenubarCB, PV_DTM_IN },
  { "DTM Out...",      'O', GUImainMenubarCB, PV_DTM_OUT },
  { "----" },
  { "Exit Program...", 'E', GUImainMenubarCB, PV_EXIT_PROGRAM },
  { NULL },
};

static XmxMenubarStruct WindowsMenu[] =
{
  { "Open Polyview...", 'O', GUImainMenubarCB, PV_OPEN_POLYVIEW },
  { "Open Palette...",  'P', GUImainMenubarCB, PV_OPEN_PALETTE },
  { "Open Info...",     'I', GUImainMenubarCB, PV_OPEN_INFO },
  { "Open Time...",     'T', GUImainMenubarCB, PV_OPEN_TIME },
#if 0
  { "----" },
  { "#Autoplace",       'A', GUImainMenubarCB, PV_AUTOPLACE },
#endif
  { NULL },
};

static XmxMenubarStruct OptionsMenu[] =
{
#ifdef DO_AUDIO
  { "#Audible",         'S', GUImainMenubarCB, PV_TOGGLE_SOUND },
  { "----" },
#endif
#ifdef __sgi
  { "#Hardware Dither", 'D', GUImainMenubarCB, PV_TOGGLE_DITHER },
#endif
  { NULL },
};

static XmxMenubarStruct HelpMenu[] =
{
  { "About",       'A', GUImainMenubarCB, PV_HELP_ABOUT },
  { "On Commands", 'C', GUImainMenubarCB, PV_HELP_ONCOMMANDS},
  { "On Window",   'W', GUImainMenubarCB, PV_MAIN_HELP_ONWINDOW },
  { NULL },
};

static XmxMenubarStruct MainMenu[] = 
{
  { "File",    'F', NULL, NULL, FileMenu },
  { "Windows", 'W', NULL, NULL, WindowsMenu },
#ifdef __sgi
  /* No options on an RS/6000, yet. */
  { "Options", 'O', NULL, NULL, OptionsMenu },
#endif
  { "Help",    'H', NULL, NULL, HelpMenu },
  { NULL },
};

/* ----------------------------- post_status ------------------------------ */
/* This is called to post the main window upon entering the GUI
   startup routines. */
int post_status(state_t *state)
{
  window_t *win;
  status_t *s;
  Widget parent_form;
  Widget command_frame;
  Widget status_frame, status_rc;

  /* This will happen ONCE only. */
  if (STA_STATUS(state) != NULL)
    return ST_ERROR;

  STA_STATUS(state) =
    win = add_window(state, "status", STATUS, init_status, redraw_status,
                     animate_status, event_status, notify_status, 
                     destroy_status);
  
  /* If the window was not created, abort the program with an error. */
  assert (win != NULL);
  
  /* Create the associated status record to hold information. */
  s = (status_t *)PVMALLOC (sizeof(status_t));
  WIN_IMAGE(win) = (void *)s;
  state->status = win;
  
  /* Clear out components that aren't created now. */
  s->load_script = NULL;

  /* Make a uniqid and set up the window_s struct somewhere in here. */
  win->id = XmxMakeNewUniqid ();
  XmxSetUniqid (win->id);
  /* In this case, we already have a base Widget. */
  win->base = state->base;

  /* Create the main form. */
  parent_form = XmxMakeForm (state->base);

  /* Make the menubar first. */
  s->menubar = XmxRMakeMenubar (parent_form, MainMenu);

#ifdef __sgi
  /* Dithering is on by default. */
  XmxRSetToggleState (s->menubar, PV_TOGGLE_DITHER, XmxSet);
#endif /* __sgi */
  
  /* Make the command area next. */
  command_frame = XmxMakeFrame (parent_form, XmxShadowOut);
  XmxSetArg (XmNwidth, 600);
  XmxSetArg (XmNheight, 204);
  s->command = XmxMakeCommand 
    (command_frame, "Enter Polyview command:", GUIcommandCB, 0);
  s->command_list = XmCommandGetChild (s->command, XmDIALOG_HISTORY_LIST);
  
  /* Make the status area last. */
  status_frame = XmxMakeFrame (parent_form, XmxShadowOut);
  XmxSetArg (XmNmarginWidth, 5);
  XmxSetArg (XmNmarginHeight, 5);
  status_rc = XmxMakeVerticalRowColumn (status_frame);
  XmxMakeLabel (status_rc, "Status:");
  XmxSetArg (XmNeditable, False);
  XmxSetArg (XmNcursorPositionVisible, False);
  s->status = XmxMakeText (status_rc);
  GUIsetStatus
    ("NCSA Polyview 3.1.  Problems -> email polyview@ncsa.uiuc.edu.");
  
  /* Constraints for parent_form. */
  XmxSetConstraints
    (s->menubar->base, XmATTACH_FORM, XmATTACH_NONE, XmATTACH_FORM, 
     XmATTACH_FORM, NULL, NULL, NULL, NULL);
  XmxSetConstraints
    (command_frame, XmATTACH_WIDGET, XmATTACH_WIDGET, XmATTACH_FORM, 
     XmATTACH_FORM, s->menubar->base, status_frame, NULL, NULL);
  XmxSetConstraints
    (status_frame, XmATTACH_NONE, XmATTACH_FORM, XmATTACH_FORM, XmATTACH_FORM,
     NULL, NULL, NULL, NULL);

  return ST_OKAY;
}
